/**********************************************************************
   Mark Butcher    Bsc (Hons) MPhil MIET

   M.J.Butcher Consulting
   Birchstrasse 20f,    CH-5406, Rtihof
   Switzerland

   www.uTasker.com    Skype: M_J_Butcher

   ---------------------------------------------------------------------
   File:        mass_storage.c
   Project:     uTasker project
   ---------------------------------------------------------------------
   Copyright (C) M.J.Butcher Consulting 2004..2010
   *********************************************************************

  05.01.2010 Modify free cluster value written to info sector during formatting {1}
  19.01.2010 Reset file length when creating files to avoid taking values from previously deleted entres {2}
  20.01.2010 Add new time when writing to a file - add creation parameter to function fnSetTimeDate() {3}
  21.01.2010 Correction in fnAllocateCluster() which could cause FAT corruption on cluster growth {4}
  26.01.2010 Add SD controller interface                                 {5}
  27.01.2010 Correct creating, renaming and deleting files and directories in referenced sub-directories with help of flag UTDIR_SET_START {6}

*/


/* =================================================================== */
/*                           include files                             */
/* =================================================================== */

#include "config.h"

//#define MJBC_MULTIPLE_WRITE_TEST
//#define MJBC_MULTIPLE_WRITE_TEST_2

#if defined SDCARD_SUPPORT

/* =================================================================== */
/*                          local definitions                          */
/* =================================================================== */

#define fnMemoryDebugMsg(x) fnDebugMsg(x)
//#define fnMemoryDebugMsg(x)


#define OWN_TASK                         TASK_MASS_STORAGE

#define T_POWER_STABILISE                (DELAY_LIMIT)(SEC * 0.05)
#define T_NEXT_CHECK                     (DELAY_LIMIT)(SEC * SD_CARD_RETRY_INTERVAL)

#define E_POWER_STABILISED               1
#define E_POLL_SD_CARD                   2

#define SD_STATE_STARTING                0
#define SD_STATE_WAIT_CARD               1
#define SD_STATE_STABILISING             2
#define SD_STATE_GO_IDLE                 3
#define SD_STATE_IF_COND                 4
#define SD_STATE_APP_CMD55_CMD41         5
#define SD_STATE_APP_CMD55_CMD41_2       6
#define SD_STATE_OCR                     7
#define SD_STATE_GET_SECTORS             8
#define SD_MOUNTING_1                    9
#define SD_MOUNTING_2                    10
#define SD_MOUNTING_3                    11
#define SD_MOUNTING_4                    12
#define SD_STATE_READY                   13

#define SD_STATE_FORMATTING_DISK_1       14
#define SD_STATE_FORMATTING_DISK_2       15
#define SD_STATE_FORMATTING_DISK_3       16
#define SD_STATE_FORMATTING_DISK_4       17
#define SD_STATE_FORMATTING_DISK_5       18
#define SD_STATE_FORMATTING_DISK_6       19
#define SD_STATE_FORMATTING_DISK_7       20
#define SD_STATE_FORMATTING_DISK_8       31
#define SD_STATE_FORMATTING_DISK_9       32
#define SD_STATE_FORMATTING_DISK_10      33
#define SD_STATE_FORMATTING_DISK_11      34

#define SD_NOT_FORMATTED                 50



// Definitions for MMC/SDC commands (0x40 + command number)
//
#ifdef SD_CONTROLLER_AVAILABLE
    #define COMMAND_OFFSET               0x00                            // no software offset since this is performed by the controller
#else
    #define COMMAND_OFFSET               0x40
#endif
#define GO_IDLE_STATE_CMD0               (COMMAND_OFFSET + 0)
#define SEND_OP_COND_CMD1                (COMMAND_OFFSET + 1)
#define SEND_IF_COND_CMD8                (COMMAND_OFFSET + 8)
#define SEND_CSD_CMD9                    (COMMAND_OFFSET + 9)
#define STOP_TRANSMISSION_CMD12          (COMMAND_OFFSET + 12)
#define SET_BLOCKLEN_CMD16               (COMMAND_OFFSET + 16)
#define READ_SINGLE_BLOCK_CMD17          (COMMAND_OFFSET + 17)
#define WRITE_BLOCK_CMD24                (COMMAND_OFFSET + 24)
#define SEND_OP_COND_ACMD_CMD41          (COMMAND_OFFSET + 41)
#define APP_CMD_CMD55                    (COMMAND_OFFSET + 55)           // defines to the card that the following command is an application specific command rather than a standard command
#define READ_OCR_CMD58                   (COMMAND_OFFSET + 58)


#define VOLTAGE_2_7__3_6                 0x01
#define CHECK_PATTERN                    0xaa
#define CS_SEND_IF_COND_CMD8             0x87

#define HIGH_CAPACITY_SD_CARD_MEMORY     0x40

#define CS_GO_IDLE_STATE_CMD0            0x95
#define CS_SET_BLOCKLEN_CMD16            0x00
#define CS_SEND_OP_COND_ACMD_CMD41       0x00
#define CS_APP_CMD_CMD55                 0x00
#define CS_READ_OCR_CMD58                0x00
#define CS_SEND_OP_COND_CMD1             0x00
#define CS_SEND_CSD_CMD9                 0x00

#define R1_IN_IDLE_STATE                 0x01
#define R1_ERASE_RESET                   0x02
#define R1_ILLEGAL_COMMAND               0x04
#define R1_COMMAND_CRC_ERROR             0x08
#define R1_ERASE_SEQUENCE_ERROR          0x10
#define R1_ADDRESS_ERROR                 0x20
#define R1_PARAMETER_ERROR               0x40


#define SDC_CARD_VERSION_2PLUS           0x01

#define SD_CARD_BUSY                     0x80


#define CARD_BUSY_WAIT                   1
#define ERROR_CARD_TIMEOUT               2
#define ERROR_SECTOR_INVALID             3


#define _IDLE_MEMORY                     0x00
#define _INITIALISED_MEMORY              0x01
#define _READING_MEMORY                  0x02
#define _WRITING_MEMORY                  0x04
#define _FORMATTING_DISK                 0x08
#define _COUNTING_CLUSTERS               0x10


#define END_DIRECTORY_ENTRIES            -2
#define MATCH_FALSE                      -1
#define MATCH_CONTINUE                   0
#define MATCH_SUCCESSFUL                 1


// File search results
//
#define FULLY_QUALIFIED_LONG_NAME        3
#define LONG_NAME_PARAGRAPH              2
#define FULLY_QUALIFIED_SHORT_NAME       1
#define SHORT_NAME_PARAGRAPH             0
#define INVALID_PARAGRAPH                -1


// FAT32 BPB_ExtFlags flags
//
#define BPB_ExtFlags_0_ZERO_BASED_NUMBER 0x0f                            // only valid if mirroring is disabled
#define BPB_ExtFlags_0_MIRRORED_FAT      0x00
#define BPB_ExtFlags_0_ONE_FAT           0x80                            // mirroring disabled

#define DIR_NAME_FREE                    0xe5

#define DIR_ATTR_READ_ONLY               0x01
#define DIR_ATTR_HIDDEN                  0x02
#define DIR_ATTR_SYSTEM                  0x04
#define DIR_ATTR_VOLUME_ID               0x08
#define DIR_ATTR_DIRECTORY               0x10
#define DIR_ATTR_ARCHIVE                 0x20
#define DIR_ATTR_LONG_NAME               (DIR_ATTR_READ_ONLY | DIR_ATTR_HIDDEN | DIR_ATTR_SYSTEM | DIR_ATTR_VOLUME_ID)
#define DIR_ATTR_MASK                    (DIR_ATTR_LONG_NAME | DIR_ATTR_DIRECTORY | DIR_ATTR_ARCHIVE)

#define BACK_SLASH                       0x5c
#define FORWARD_SLASH                    0x2f

#define NT_FLAG                          0x18

#define NEW_ABSOLUTE_CLUSTER             0x00
#define INITIALISE_DIR_CLUSTER           0x01
#define INITIALISE_DIR_EXTENSION         0x02
#define UPDATE_FAT_END                   0x04
#define UPDATE_FAT_END_IN_DIFF_SECT      0x08
#define NEW_RELATIVE_CLUSTER             (UPDATE_FAT_END | UPDATE_FAT_END_IN_DIFF_SECT)


#define _CHAR_REJECT                     0x01
#define _CHAR_TERMINATE                  0x02
#define _CHAR_CAPITAL                    0x04
#define _CHAR_SMALL                      0x08
#define _CHAR_NUMBER                     0x10
#define _CHAR_REJECT_NON_JAP             0x20

/* =================================================================== */
/*                      local structure definitions                    */
/* =================================================================== */


/* =================================================================== */
/*                 local function prototype declarations               */
/* =================================================================== */

static int  utReadDisk(UTDISK *ptr_utDisk, unsigned long ulSectorNumber, unsigned char ucSectorCount);
static int  utReadDiskData(UTDISK *ptr_utDisk, unsigned long ulSector, void *ptrBuf, unsigned short usOffset, unsigned short usLength);
static int  fnLoadSector(UTDISK *ptr_utDisk, unsigned long ulSector);
static int  fnSendSD_command(const unsigned char ucCommand[6], unsigned char *ucResult, unsigned char *ptrReturnData);
static const unsigned char *fnCreateCommand(unsigned char ucCommand, unsigned long ulValue);
static int  fnGetSector(unsigned char *ptrBuf);
static int  fnReadPartialSector(unsigned char *ptrBuf, unsigned short usStart, unsigned short usStop);
static int  fnWaitSD_ready(int iMaxWait);
static int  ut_read_disk(UTDISK *ptr_utDisk, unsigned char ucReadSize);
static void fnCardNotFormatted(void);
static void fnInitialisationError(void);
#ifdef UTFAT_WRITE
    static int  utCommitSector(UTDISK *ptr_utDisk, unsigned long ulSector);
    static int  utCommitSectorData(UTDISK *ptr_utDisk, void *ptrBuffer, unsigned long ulSectorNumber);
    static int  utDeleteSector(UTDISK *ptr_utDisk, unsigned long ulSectorNumber);
    static unsigned long fnAllocateCluster(UTDISK *ptr_utDisk, unsigned long ulPresentCluster, unsigned char ucClusterType);
    static int  fnDeleteFileContent(UTFILE *ptr_utFile, UTDISK *ptr_utDisk, int iDestroyClusters);
    static int  fnDeleteClusterChain(unsigned long ulClusterStart, unsigned char ucDrive, int iDestroyClusters);
    static void fnAddInfoSect(INFO_SECTOR_FAT32 *ptrInfoSector, unsigned long ulFreeCount, unsigned long ulNextFree);
#endif

static int fnFileLocked(UTFILE *ptr_utFile);


/* =================================================================== */
/*                             constants                               */
/* =================================================================== */

static const unsigned char ucGO_IDLE_STATE_CMD0[6]      = {GO_IDLE_STATE_CMD0, 0x00, 0x00, 0x00, 0x00, CS_GO_IDLE_STATE_CMD0};
static const unsigned char ucIF_COND_CMD8[6]            = {SEND_IF_COND_CMD8, 0x00, 0x00, VOLTAGE_2_7__3_6, CHECK_PATTERN, CS_SEND_IF_COND_CMD8};
static const unsigned char ucSEND_CSD_CMD9[6]           = {SEND_CSD_CMD9, 0x00, 0x00, 0x00, 0x00, CS_SEND_CSD_CMD9};
static const unsigned char ucAPP_CMD_CMD55[6]           = {APP_CMD_CMD55, 0x00, 0x00, 0x00, 0x00, CS_APP_CMD_CMD55};
static const unsigned char ucSEND_OP_COND_ACMD_CMD41[6] = {SEND_OP_COND_ACMD_CMD41, HIGH_CAPACITY_SD_CARD_MEMORY, 0x00, 0x00, 0x00, CS_SEND_OP_COND_ACMD_CMD41};
static const unsigned char ucREAD_OCR_CMD58[6]          = {READ_OCR_CMD58, 0x00, 0x00, 0x00, 0x00, CS_READ_OCR_CMD58};

static unsigned char ucEmptyFAT32[12] = {
    LITTLE_LONG_WORD_BYTES(MEDIA_VALUE_FIXED),
    LITTLE_LONG_WORD_BYTES(0xffffffff),
    LITTLE_LONG_WORD_BYTES(CLUSTER_MASK)
};


static const unsigned char ucCharacterTable[] = {
    (0),                                                                 // !
    (_CHAR_REJECT),                                                      // "
    (0),                                                                 // #
    (0),                                                                 // $
    (0),                                                                 // %
    (0),                                                                 // &
    (0),                                                                 // 
    (0),                                                                 // (
    (0),                                                                 // )
    (_CHAR_REJECT),                                                      // *
    (_CHAR_REJECT),                                                      // +
    (0),                                                                 // ,
    (0),                                                                 // -
    (0),                                                                 // .
    (_CHAR_TERMINATE),                                                   // /
    (_CHAR_NUMBER),                                                      // 0
    (_CHAR_NUMBER),                                                      // 1
    (_CHAR_NUMBER),                                                      // 2
    (_CHAR_NUMBER),                                                      // 3
    (_CHAR_NUMBER),                                                      // 4
    (_CHAR_NUMBER),                                                      // 5
    (_CHAR_NUMBER),                                                      // 6
    (_CHAR_NUMBER),                                                      // 7
    (_CHAR_NUMBER),                                                      // 8
    (_CHAR_NUMBER),                                                      // 9
    (_CHAR_REJECT),                                                      // :
    (_CHAR_REJECT),                                                      // ;
    (_CHAR_REJECT),                                                      // <
    (_CHAR_REJECT),                                                      // =
    (_CHAR_REJECT),                                                      // >
    (0),                                                                 // ?
    (0),                                                                 // @
    (_CHAR_CAPITAL),                                                     // A
    (_CHAR_CAPITAL),                                                     // B
    (_CHAR_CAPITAL),                                                     // C
    (_CHAR_CAPITAL),                                                     // D
    (_CHAR_CAPITAL),                                                     // E
    (_CHAR_CAPITAL),                                                     // F
    (_CHAR_CAPITAL),                                                     // G
    (_CHAR_CAPITAL),                                                     // H
    (_CHAR_CAPITAL),                                                     // I
    (_CHAR_CAPITAL),                                                     // J
    (_CHAR_CAPITAL),                                                     // K
    (_CHAR_CAPITAL),                                                     // L
    (_CHAR_CAPITAL),                                                     // M
    (_CHAR_CAPITAL),                                                     // N
    (_CHAR_CAPITAL),                                                     // O
    (_CHAR_CAPITAL),                                                     // P
    (_CHAR_CAPITAL),                                                     // Q
    (_CHAR_CAPITAL),                                                     // R
    (_CHAR_CAPITAL),                                                     // S
    (_CHAR_CAPITAL),                                                     // T
    (_CHAR_CAPITAL),                                                     // U
    (_CHAR_CAPITAL),                                                     // V
    (_CHAR_CAPITAL),                                                     // W
    (_CHAR_CAPITAL),                                                     // X
    (_CHAR_CAPITAL),                                                     // Y
    (_CHAR_CAPITAL),                                                     // Z
    (_CHAR_REJECT_NON_JAP),                                              // [
    (_CHAR_TERMINATE),                                                   // back slash
    (_CHAR_REJECT_NON_JAP),                                              // ]
    (0),                                                                 // ^
    (0),                                                                 // _
    (0),                                                                 // '
    (_CHAR_SMALL),                                                       // a
    (_CHAR_SMALL),                                                       // b
    (_CHAR_SMALL),                                                       // c
    (_CHAR_SMALL),                                                       // d
    (_CHAR_SMALL),                                                       // e
    (_CHAR_SMALL),                                                       // f
    (_CHAR_SMALL),                                                       // g
    (_CHAR_SMALL),                                                       // h
    (_CHAR_SMALL),                                                       // i
    (_CHAR_SMALL),                                                       // j
    (_CHAR_SMALL),                                                       // k
    (_CHAR_SMALL),                                                       // l
    (_CHAR_SMALL),                                                       // m
    (_CHAR_SMALL),                                                       // n
    (_CHAR_SMALL),                                                       // o
    (_CHAR_SMALL),                                                       // p
    (_CHAR_SMALL),                                                       // q
    (_CHAR_SMALL),                                                       // r
    (_CHAR_SMALL),                                                       // s
    (_CHAR_SMALL),                                                       // t
    (_CHAR_SMALL),                                                       // u
    (_CHAR_SMALL),                                                       // v
    (_CHAR_SMALL),                                                       // w
    (_CHAR_SMALL),                                                       // x
    (_CHAR_SMALL),                                                       // y
    (_CHAR_SMALL),                                                       // z
    (0),                                                                 // {
    (_CHAR_REJECT_NON_JAP),                                              // |
    (0),                                                                 // }
    (0),                                                                 // ~
};


/* =================================================================== */
/*                     global variable definitions                     */
/* =================================================================== */


/* =================================================================== */
/*                      local variable definitions                     */
/* =================================================================== */


static UTDISK utDisks[SDCARDS] = {{0}};

static int CommandTimeout = 5000;                                        // change depending on SPI speed

static int iMemoryOperation = _IDLE_MEMORY;
static int iMemoryState = SD_STATE_STARTING;

static unsigned long ulClusterSectorCheck = 0;
static unsigned long ulActiveFreeClusterCount = 0;
static UTASK_TASK cluster_task = 0;



///////////////////////// multiple FLASH write development/////////////////////////////////////


#define MANAGED_READ                0x00
#define MANAGED_WRITE               0x01
#define MANAGED_LOCK                0x02
#define MANAGED_DELETE              0x04
#define MANAGED_FLUSH_FILE_SYSTEM   0x10
#define WAITING_DELETE              0x20
#define WAITING_READ                0x40
#define WAITING_WRITE               0x80

#define MANAGED_FILE_NO_ACCESS      -1
#define MANAGED_FILE_NO_FILE_HANDLE -2

#if defined UTMANAGED_FILE_COUNT && UTMANAGED_FILE_COUNT > 0


typedef struct stUTMANAGED_FILE
{
    unsigned char   managed_mode;
    unsigned char   managed_owner;
    UTFILE         *utManagedFile;
}   UTMANAGED_FILE;

static UTMANAGED_FILE utManagedFiles[UTMANAGED_FILE_COUNT] = {{0}};

#endif




#ifdef MANAGED_FILES
#ifdef SUB_FILE_SIZE
    #define SUBFILE_WRITE  , ucSubFileInProgress
    #define SUB_FILE_ON    ,SUB_FILE_TYPE
#else
    #define SUBFILE_WRITE
    #define SUB_FILE_ON
#endif



#define MANAGED_FILE_COUNT          5


// Managed file object
//
typedef struct stMANAGED_FILE
{
    unsigned char  *managed_list;
    unsigned char  *managed_write;
    MAX_FILE_LENGTH managed_write_length;
    MAX_FILE_LENGTH managed_size;
    UTASK_TASK      managed_owner;
    unsigned char   managed_mode;
}   MANAGED_FILE;



static MANAGED_FILE managed_files[MANAGED_FILE_COUNT] = {{0}};



extern int uOpenManagedFile(CHAR *ptrFileName, UTASK_TASK owner_task, unsigned char ucMode)
{
    int i = 0;
    int iFree = 0;
    unsigned char *ptrFile;
    ptrFile = uOpenFile(ptrFileName);                                    // get the file to be opened
    while (i < MANAGED_FILE_COUNT) {                                     // check whether this file is presently being protected
        if (managed_files[i].managed_owner != 0) {
            if (managed_files[i].managed_mode & (MANAGED_WRITE | MANAGED_LOCK)) {
                if ((ptrFile >= managed_files[i].managed_list) && (ptrFile < (managed_files[i].managed_list + managed_files[i].managed_size))) {
                    return MANAGED_FILE_NO_ACCESS;                       // this file can presently not be accessed
                }
            }
        }
        else {
            iFree = (i + 1);                                             // free space found
        }
        i++;
    }
    if (iFree-- == 0) {
        return MANAGED_FILE_NO_FILE_HANDLE;                              // all managed file spaces are presently occupied
    }
    managed_files[iFree].managed_owner = owner_task;
    managed_files[iFree].managed_list = ptrFile;
    managed_files[iFree].managed_mode = ucMode;
    managed_files[iFree].managed_size = uGetFileLength(ptrFile);
    if (ucMode & MANAGED_WRITE) {
        if (managed_files[iFree].managed_size != 0) {
            uFileErase(ptrFile, managed_files[iFree].managed_size);      // erase the file to make space for new upload
        }
        managed_files[iFree].managed_size = SINGLE_FILE_SIZE;            // file presently occupies only one file block, but it could grow
        managed_files[iFree].managed_write_length = FILE_HEADER;
        managed_files[iFree].managed_write = ptrFile;
    }
    return iFree;                                                        // return file handle
}


#define MEDIA_BUSY        -1
#define MEDIA_FATAL_ERROR -2
static int iManagedMedia = 0;
extern int uFileManagedWrite(int fileHandle, unsigned char *ptrData, MAX_FILE_LENGTH DataLength)
{
    int iMediaState;
    if ((managed_files[fileHandle].managed_write_length + DataLength) > managed_files[fileHandle].managed_size) {
        managed_files[fileHandle].managed_size += SINGLE_FILE_SIZE;
        // delete more space!!!! but check end of file system...
        // possibly limit data length
    }
    iMediaState = fnWriteBytesFlash((managed_files[fileHandle].managed_list + managed_files[fileHandle].managed_write_length), ptrData, DataLength); // we are sure that we have enough space to save the present data so save it
    if (iMediaState < 0) {
        if (iMediaState == MEDIA_BUSY) {
            uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);              // set to polling mode
            iManagedMedia = 1;                                           // mark that this media type needs handling
            managed_files[fileHandle].managed_mode |= WAITING_WRITE;     // mark that this owner is waiting to complete a write
        }
        return iMediaState;                                              // return the reason for not being able to write at the moment
    }
    managed_files[fileHandle].managed_write_length += DataLength;
    return managed_files[fileHandle].managed_write_length;               // return new length
}

extern int uFileManagedDelete(int fileHandle)
{
    if (managed_files[fileHandle].managed_mode & MANAGED_FLUSH_FILE_SYSTEM) {// complete file system delete
        managed_files[fileHandle].managed_mode = WAITING_DELETE;
        managed_files[fileHandle].managed_size = FILE_SYSTEM_SIZE;
        managed_files[fileHandle].managed_write = uFILE_SYSTEM_START;
    }
    uFileErase(managed_files[fileHandle].managed_write, FLASH_GRANULARITY);
    managed_files[fileHandle].managed_write += FLASH_GRANULARITY;
    if (managed_files[fileHandle].managed_size <= FLASH_GRANULARITY) {
        managed_files[fileHandle].managed_mode = 0;
        return 0;                                                    // complete
    }
    else {
        managed_files[fileHandle].managed_size -= FLASH_GRANULARITY;
    }
    iManagedMedia = 1;                                               // mark that this media type needs handling
    uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);                  // set to polling mode
    return 1;                                                        // delete not yet complete
}

static void fnManagedMediaCheck(void)
{
/*
    int i = MANAGED_FILE_COUNT;
    int iMediaState;
    while (i > 0) {
        i--;
        if (managed_files[i].managed_owner != 0) {
            if (managed_files[i].managed_mode & WAITING_WRITE) {
                iMediaState = fnWriteBytesFlash((managed_files[i].managed_list + managed_files[i].managed_write_length), 0, 0); // check the present state
                if (iMediaState < 0) {
                    if (iMediaState == MEDIA_BUSY) {
                        uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);  // set to polling mode
                        iManagedMedia = 1;                               // mark that this media type needs handling
                    }
                }
                else {
                    managed_files[i].managed_mode &= ~WAITING_WRITE;     // remove waiting flag
                    // inform waiting task that the write is now possible
                }
            }
            else if (managed_files[i].managed_mode & WAITING_DELETE) {
                if (uFileManagedDelete(i) == 0) {
                    // delete complete - inform owner
                    if (managed_files[i].managed_owner == TASK_FTP) {    // special case for FTP since it is not a task
                        extern void fnFTP_managed_file(int i, unsigned char ucEvent);
                        #define DELETE_TERMINATED 0x01
                        fnFTP_managed_file(i, DELETE_TERMINATED);
                    }
                }
            }
        }
    }
    */
}


MAX_FILE_LENGTH uFileManagedClose(int fileHandle, unsigned char ucMimeType)
{
    unsigned char ucMimeTypeToSave = ucMimeType;
  //MAX_FILE_LENGTH fileLength = uFileCloseMime((managed_list[fileHandle] + managed_write_length[fileHandle]), &ucMimeTypeToSave);
    //fnWriteBytesFlash((FileInProgress + LengthInProgress++), &ucEndByte, 1); // add last byte to end of file
    fnWriteBytesFlash((managed_files[fileHandle].managed_list + sizeof(MAX_FILE_LENGTH)), &ucMimeTypeToSave, 1);
    managed_files[fileHandle].managed_write_length -= FILE_HEADER;
    fnWriteBytesFlash(managed_files[fileHandle].managed_list, (unsigned char *)&managed_files[fileHandle].managed_write_length, sizeof(MAX_FILE_LENGTH));// write the length to the file
    managed_files[fileHandle].managed_owner = 0;                         // the file is now closed
    return managed_files[fileHandle].managed_write_length;
}


static unsigned char ucData[10] = {0,1,2,3,4,5,6,7,8,9};                 // test file content

static void fnTestMultipleWrites(void)
{
    unsigned char ucMimeType1, ucMimeType2;
    MAX_FILE_LENGTH lengthFile1, lengthFile2;

    int fileHandle1 = uOpenManagedFile((CHAR *)"0.htm", OWN_TASK, MANAGED_WRITE);
    int fileHandle2 = uOpenManagedFile((CHAR *)"1.htm", OWN_TASK, MANAGED_WRITE);
    unsigned char *ptrFile1 = uOpenFile((CHAR *)"0.htm");
    unsigned char *ptrFile2 = uOpenFile((CHAR *)"1.htm");
    uFileManagedWrite(fileHandle1, ucData, sizeof(ucData));
    uFileManagedWrite(fileHandle2, ucData, sizeof(ucData));
    uFileManagedClose(fileHandle1, MIME_HTML);
    uFileManagedClose(fileHandle2, MIME_HTML);

    fnGetParsFile((ptrFile1 + sizeof(MAX_FILE_LENGTH)), &ucMimeType1, 1); // check compatibility
    fnGetParsFile((ptrFile2 + sizeof(MAX_FILE_LENGTH)), &ucMimeType2, 1);
    lengthFile1 = uGetFileLength(ptrFile1);
    lengthFile2 = uGetFileLength(ptrFile2);
}
#endif


///////////////////////// END multiple FLASH write development/////////////////////////////////////

























// Mass storage task
//
extern void fnMassStorage(TTASKTABLE *ptrTaskTable)
{
#ifdef _WINDOWS
    static int iFormatCount;
#endif
    int iActionResult = 0;
    unsigned char ucInputMessage[HEADER_LENGTH];                         // reserve space for receiving messages
    unsigned char ucResult = 0;
    unsigned char ucData[18];

#ifdef MJBC_MULTIPLE_WRITE_TEST
    if (iManagedMedia != 0) {
        iManagedMedia = 0;
        fnManagedMediaCheck();
    }
#endif
    if (iMemoryOperation & _READING_MEMORY) {                            // reading
        if ((iActionResult = utReadDisk(&utDisks[0], 0, 0)) == CARD_BUSY_WAIT) {
            return;                                                      // still reading so keep waiting
        }
    }    
    if (!(iMemoryOperation & _INITIALISED_MEMORY)) {                     // if initialisation in progress
        static unsigned long ulFatSectors;
        static unsigned long ulFAT32size;
        static unsigned char ucFatCopies;
        switch (iMemoryState) {                                          // perform the initialisation state-event process
        case SD_STATE_STARTING:
        case SD_STATE_WAIT_CARD:
#ifdef MJBC_MULTIPLE_WRITE_TEST_2
            fnTestMultipleWrites();
#endif
#ifdef SD_CONTROLLER_AVAILABLE                                           // {5}
            POWER_UP(PCSDC);                                             // power up the SC controller interface
            PINSEL1 |= (PINSEL1_P0_22_MCIDATA0 | PINSEL1_P0_21_MCIPWR | PINSEL1_P0_20_MCIMD | PINSEL1_P0_19_MCICLK);
            PINSEL4 |= (PINSEL4_P2_13_MCIDAT3 | PINSEL4_P2_12_MCIDAT2 | PINSEL4_P2_11_MCIDAT1);
            SCS |= MCIPWR;                                               // configure polarity of power control signal
            MCICommand = 0;
            MCIDataCtrl = 0;
	        MCIClear = (CmdCrcFailClr | DataCrcFailClr | CmdTimeOutClr | DataTimeOutClr | TxUnderrunClr | RxOverrunClr | CmdRespEndClr | CmdSentClr | DataEndClr | StartBitErrClr | DataBlockEndClr); // clear all status flags
            MCIPower = MCI_PWR_Ctrl_PowerUp;
            iMemoryState = SD_STATE_STABILISING;                         // move to stabilisation delay state
            uTaskerMonoTimer(OWN_TASK, T_POWER_STABILISE, E_POWER_STABILISED); // wait until SD card power stabilised
            MCIClock = (MCI_CLK_PwrSave | MCI_CLK_Enable | ((PCLK/4/2/300000 - 1))); // 300kHz initial speed in 1-bit mode
#else
            INITIALISE_SPI_SD_INTERFACE();                               // initialise the SPI interface to the card
            SET_SD_DI_CS_HIGH();                                         // prepare chip select and DI ready for the initialisation
            POWER_UP_SD_CARD();                                          // power up if necessary
            iMemoryState = SD_STATE_STABILISING;                         // move to stabilisation delay state
            ENABLE_SPI_SD_OPERATION();                                   // dummy to solve an AVR32 GCC optimising problem
            uTaskerMonoTimer(OWN_TASK, T_POWER_STABILISE, E_POWER_STABILISED); // wait until SD card power stabilised
#endif
            break;

        case SD_STATE_STABILISING:                                       // delay after applying power to the SD-card
#ifdef SD_CONTROLLER_AVAILABLE                                           // {5}
            MCIPower = MCI_PWR_Ctrl_PowerOn;
            iMemoryState = SD_STATE_GO_IDLE;                             // prepare to communicate
            uTaskerMonoTimer(OWN_TASK, T_POWER_STABILISE, E_POWER_STABILISED); // wait until SD card power stabilised
            break;
#else
            {                                                            // send at least 74 clocks to the SD-card
                int i = 10;
                while (i--) {                                            // set the SD card to native command mode by sending 80 clocks
                    WRITE_SPI_CMD(0xff);                                 // write dummy tx
                    WAIT_TRANSMISSON_END();                              // wait until transmission complete
                    READ_SPI_DATA();                                     // read 10 dummy bytes from the interface in order to generate 80 clock pulses on the interface (at least 74 needed)
                }
                SET_SD_CS_LOW();                                         // assert the chip select line to the SD-card ready to start the initialisation sequence
                iMemoryState = SD_STATE_GO_IDLE;                         // move to next state
                SET_SD_CARD_MODE();                                      // allow final mode in case the DIN line had to be pulled up during native mode sequence
            }                                                            // fall through
#endif
        case SD_STATE_GO_IDLE:
            if ((iActionResult = fnSendSD_command(ucGO_IDLE_STATE_CMD0, &ucResult, 0)) != UTFAT_SUCCESS) {
                if (iActionResult == CARD_BUSY_WAIT) {
                    return;                                              // read is taking time to complete so quit for the moment
                }
                ucResult = 0;                                            // set error since the result is not as expected
            }
            if (R1_IN_IDLE_STATE != ucResult) {                          // no valid card detected so disable power and try again after a delay
                fnInitialisationError();                                 // the card is no present or is behaving incorrectly - stop and try again later
                break;
            }                                                            // SD card must return the idle state to be able to continue
                                                                         // in the idle state the card accepts only commands 0, 1, ACMD41 and 58
            iMemoryState = SD_STATE_IF_COND;                             // fall through
        case SD_STATE_IF_COND:                                           // it is mandatory for a host compliant to Physical Spec. Version 2.00 to send the CMD8 to retrieve non-supported voltage range
            if ((iActionResult = fnSendSD_command(ucIF_COND_CMD8, &ucResult, ucData)) != UTFAT_SUCCESS) {
                if (iActionResult == CARD_BUSY_WAIT) {
                    return;                                              // read is taking time to complete so quit for the moment
                }
                fnInitialisationError();                                 // the card is behaving incorrectly - stop and try again later
                break;
            }
            if (ucResult == SDC_CARD_VERSION_2PLUS) {                    // version 2 or higher
                if ((ucData[2] == VOLTAGE_2_7__3_6) && (ucData[3] == CHECK_PATTERN)) { // check whether the card can operate between 2.7V and 3.6V
                    iMemoryState = SD_STATE_APP_CMD55_CMD41;             // now poll the SD card until it accept the application command 42
                    uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);      // run again
                }
            }
            else {                                                       // version 1 or MMC type
                fnMemoryDebugMsg("SD-card V1 or MMC - not supported!\r\n");
                SET_SD_CS_HIGH();
                POWER_DOWN_SD_CARD();
                iMemoryState = SD_STATE_WAIT_CARD;
                uTaskerMonoTimer(OWN_TASK, T_NEXT_CHECK, E_POLL_SD_CARD);// try again later
            }
            break;
        case SD_STATE_APP_CMD55_CMD41:
            if ((iActionResult = fnSendSD_command(ucAPP_CMD_CMD55, &ucResult, 0)) != UTFAT_SUCCESS) {
                if (iActionResult == CARD_BUSY_WAIT) {
                    return;                                              // read is taking time to complete so quit for the moment
                }
                fnInitialisationError();                                 // the card is behaving incorrectly - stop and try again later
                break;
            }
            if (ucResult > R1_IN_IDLE_STATE) {
                uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);          // try again
                break;
            }
            iMemoryState = SD_STATE_APP_CMD55_CMD41_2;                   // fall through to send the application command
        case SD_STATE_APP_CMD55_CMD41_2:
            if ((iActionResult = fnSendSD_command(ucSEND_OP_COND_ACMD_CMD41, &ucResult, 0)) != UTFAT_SUCCESS) {
                if (iActionResult == CARD_BUSY_WAIT) {
                    return;                                              // read is taking time to complete so quit for the moment
                }
                fnInitialisationError();                                 // the card is behaving incorrectly - stop and try again later
                break;
            }
            if (ucResult != 0) {
                iMemoryState = SD_STATE_APP_CMD55_CMD41;                 // loop back
                uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);          // try again
                break;
            }
            iMemoryState = SD_STATE_OCR;                                 // fall through to issue CMD58
        case SD_STATE_OCR:
            if ((iActionResult = fnSendSD_command(ucREAD_OCR_CMD58, &ucResult, ucData)) != UTFAT_SUCCESS) { // get card capacity information
                if (iActionResult == CARD_BUSY_WAIT) {
                    return;                                              // read is taking time to complete so quit for the moment
                }
                fnInitialisationError();                                 // the card is behaving incorrectly - stop and try again later
                break;
            }
            fnMemoryDebugMsg("SD-card V2 - ");
            if (ucData[0] & HIGH_CAPACITY_SD_CARD_MEMORY) {              // check the CCS bit
                utDisks[0].ucDiskFlags = HIGH_CAPACITY_SD_CARD;
                fnMemoryDebugMsg("High Capacity\r\n");
            }
            else {
                fnMemoryDebugMsg("standard\r\n");
            }
            iMemoryState = SD_STATE_GET_SECTORS;                         // fall through to issue CMD9 and read card specific data
        case SD_STATE_GET_SECTORS:
            if ((iActionResult = fnSendSD_command(ucSEND_CSD_CMD9, &ucResult, ucData)) != UTFAT_SUCCESS) { // get card capacity information
                if (iActionResult == CARD_BUSY_WAIT) {
                    return;                                              // read is taking time to complete so quit for the moment
                }
                fnInitialisationError();                                 // the card is behaving incorrectly - stop and try again later
                break;
            }
            uMemcpy(utDisks[0].utFileInfo.ucCardSpecificData, &ucData[2], 16); // back up the card specific data since it can be interesting later
            if (ucData[2] & HIGH_CAPACITY_SD_CARD_MEMORY) {              // high capacity
                utDisks[0].ulSD_sectors = (((ucData[9] << 16) + (ucData[10] << 8)  + ucData[11]) + 1);// SD version 2 assumed
                utDisks[0].ulSD_sectors *= 1024;                         // the number of 512 byte sectors on the SD card
            }
            else {                                                       // standard capacity
                utDisks[0].ulSD_sectors = ((((ucData[8] & 0x03) << 10) + (ucData[9] << 2) + (ucData[10] >> 6)) + 1);// SD version 2 assumed
                utDisks[0].ulSD_sectors *= (1 << ((((ucData[11] & 0x03) << 1) + (ucData[12] >> 7)) + 2)); // the number of 512 byte sectors on the SD card
                if ((ucData[7] & 0x0f) == 0x0a) {                        // 1024 byte block length indicates 2G card
                    utDisks[0].ulSD_sectors *= 2;
                }
            }
            fnMemoryDebugMsg("\r\n");
            if (utDisks[0].ptrSectorData == 0) {
    #if defined _ALIGN_HEAP_4
                utDisks[0].ptrSectorData = uMalloc(512);                  // allocate a buffer which will contain data read from the present sector
    #else
                utDisks[0].ptrSectorData = uMallocAlign(512, sizeof(unsigned long)); // allocate a buffer which will contain data read from the present sector - ensure aligned to long word boundary!
    #endif
            }
            iMemoryState = SD_MOUNTING_1;
            uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);              // schedule again to attempt mounting
            utDisks[0].ucDriveNumber = 0;
            SET_SD_CS_HIGH();
            SET_SPI_SD_INTERFACE_FULL_SPEED();                           // speed up the SPI interface since initialisation is complete
            break;
        case SD_MOUNTING_1:                                              // start mounting
            utDisks[0].ulPresentSector = 0;                              // the first sector that will be accessed
            if ((iActionResult = ut_read_disk(&utDisks[0], 1)) ==  CARD_BUSY_WAIT) { // read boot sector from disk
                return;                                                  // read is taking time to complete so quit for the moment
            }
            iMemoryState = SD_MOUNTING_2;
        case SD_MOUNTING_2:
            if (ERROR_SECTOR_INVALID == iActionResult) {
                fnCardNotFormatted();
                break;
            }
            else {
                iMemoryState = SD_MOUNTING_3;
                if (iActionResult != 0) {
                    fnInitialisationError();                             // the card is behaving incorrectly - stop and try again later
                    break;
                }
                else {
                    EXTENDED_BOOT_RECORD *ptrExtendedBootSector = (EXTENDED_BOOT_RECORD *)utDisks[0].ptrSectorData; // check to see whether partitions are defined rathern than it being a boot record
                    BOOT_SECTOR_FAT32 *ptrBootSector = (BOOT_SECTOR_FAT32 *)utDisks[0].ptrSectorData;
                    if ((ptrExtendedBootSector->EBR_partition_table[0].partition_type != 0) && (ptrBootSector->BS_FilSysType[0] != 'F')) {
                        utDisks[0].ulPresentSector = ((ptrExtendedBootSector->EBR_partition_table[0].start_sector[3] << 24) + (ptrExtendedBootSector->EBR_partition_table[0].start_sector[2] << 16) + (ptrExtendedBootSector->EBR_partition_table[0].start_sector[1] << 8) + ptrExtendedBootSector->EBR_partition_table[0].start_sector[0]);
                        if ((iActionResult = ut_read_disk(&utDisks[0], 1)) ==  CARD_BUSY_WAIT) { // read boot sector from disk
                            return;                                      // read is taking time to complete so quit for the moment
                        }
                    }
                    else {
                        utDisks[0].ulPresentSector = 0;                  // the boot sector is 0
                    }
                }
            }                                                            // fall-through
        case SD_MOUNTING_3:
            if (ERROR_SECTOR_INVALID == iActionResult) {
                fnCardNotFormatted();
                break;
            }
            else {
                BOOT_SECTOR_FAT32 *ptrBootSector;
                unsigned long  ulTotalSections;
                unsigned long  ulFatSize;
                unsigned long  ulFirstDataSector;
                unsigned long  ulCountofClusters;
                unsigned short BPB_RsvdSecCnt;
                unsigned short BPB_FSInfo;
                unsigned short BPB_BytesPerSec;
                unsigned char  BPB_SecPerClus;

                ptrBootSector = (BOOT_SECTOR_FAT32 *)utDisks[0].ptrSectorData; // the buffer content is a boot sector
                BPB_BytesPerSec = (ptrBootSector->boot_sector_bpb.BPB_BytesPerSec[1] << 8);
                ulFatSize = ((ptrBootSector->boot_sector_bpb.BPB_FATSz16[1] << 8) + ptrBootSector->boot_sector_bpb.BPB_FATSz16[0]);
                if (ulFatSize == 0) {                                    // FAT32 will indicate the size in the BPB_FATSz32 field instead
                    ulFatSize = ((ptrBootSector->BPB_FATSz32[3] << 24) + (ptrBootSector->BPB_FATSz32[2] << 16) + (ptrBootSector->BPB_FATSz32[1] << 8) + ptrBootSector->BPB_FATSz32[0]);
                }
                ulTotalSections = ((ptrBootSector->boot_sector_bpb.BPB_TotSec16[1] << 8) + ptrBootSector->boot_sector_bpb.BPB_TotSec16[0]);
                if (ulTotalSections == 0) {                              // FAT32 will indicate the size in the BPB_FATSz32 field instead
                    ulTotalSections = ((ptrBootSector->boot_sector_bpb.BPB_TotSec32[3] << 24) + (ptrBootSector->boot_sector_bpb.BPB_TotSec32[2] << 16) + (ptrBootSector->boot_sector_bpb.BPB_TotSec32[1] << 8) + ptrBootSector->boot_sector_bpb.BPB_TotSec32[0]);
                }
                utDisks[0].utFAT.ulFatSize = ulFatSize;                  // the sectors in a single FAT
                utDisks[0].utFAT.ucNumberOfFATs = ptrBootSector->boot_sector_bpb.BPB_NumFATs; // the number of FAT copies
                ulFatSize *= ptrBootSector->boot_sector_bpb.BPB_NumFATs; // the complete number of sections occupied by all FATs
                BPB_SecPerClus = ptrBootSector->boot_sector_bpb.BPB_SecPerClus;
                utDisks[0].utFAT.ucSectorsPerCluster = BPB_SecPerClus;
                BPB_RsvdSecCnt = ((ptrBootSector->boot_sector_bpb.BPB_RsvdSecCnt[1] << 8) + ptrBootSector->boot_sector_bpb.BPB_RsvdSecCnt[0]);
                utDisks[0].utFAT.ulFAT_start = (utDisks[0].ulPresentSector + BPB_RsvdSecCnt); // the boot sector plus reserved sectors
                if ((BPB_BytesPerSec < 512) || (BPB_SecPerClus == 0) || ((BPB_SecPerClus * BPB_BytesPerSec) > 32*1024)) {
                    fnMemoryDebugMsg("Malformed - ");
                    fnCardNotFormatted();
                    break;
                }
                utDisks[0].ucDiskFlags &= ~DISK_UNFORMATTED;
                utDisks[0].ucDiskFlags |= DISK_FORMATTED;
                utDisks[0].utFAT.usBytesPerSector = BPB_BytesPerSec;
                ulFirstDataSector = (BPB_RsvdSecCnt + ulFatSize);
                ulCountofClusters = ((ulTotalSections - ulFirstDataSector)/ptrBootSector->boot_sector_bpb.BPB_SecPerClus);
                if (ulCountofClusters < 65525) {                         // not FAT 32 so not supported
                    fnMemoryDebugMsg("NOT FAT32- ");
                    fnCardNotFormatted();
                    break;
                }
                utDisks[0].utFAT.ulClusterCount = ((ulTotalSections - BPB_RsvdSecCnt - ulFatSize)/BPB_SecPerClus); // total cluster count
                utDisks[0].ulLogicalBaseAddress = (utDisks[0].utFAT.ulFAT_start + ulFatSize); // data start sector (logical block address)
                BPB_FSInfo = ((ptrBootSector->BPB_FSInfo[1] << 8) + ptrBootSector->BPB_FSInfo[0]);
                utDisks[0].ulDirectoryBase = ((ptrBootSector->BPB_RootClus[3] << 24) + (ptrBootSector->BPB_RootClus[2] << 16) + (ptrBootSector->BPB_RootClus[1] << 8) + ptrBootSector->BPB_RootClus[0]); // root directory start cluster
                utDisks[0].ulVirtualBaseAddress = utDisks[0].ulLogicalBaseAddress - (utDisks[0].ulDirectoryBase * utDisks[0].utFAT.ucSectorsPerCluster);
                utDisks[0].ulPresentSector = utDisks[0].utFileInfo.ulInfoSector = (utDisks[0].ulPresentSector + BPB_FSInfo); // sector location of structure
                iMemoryState = SD_MOUNTING_4;
                if ((iActionResult = ut_read_disk(&utDisks[0], 1)) ==  CARD_BUSY_WAIT) { // read boot sector from disk
                    return;                                              // read is taking time to complete so quit for the moment
                }
            }                                                            // fall through
        case SD_MOUNTING_4:                                              // optional step for FAT32 SD-cards
            if (iActionResult == 0) {                                    // if the information sector was valid
                INFO_SECTOR_FAT32 *ptrInfo = (INFO_SECTOR_FAT32 *)utDisks[0].ptrSectorData;
                if ((ptrInfo->FSI_TrailSig[3] == 0xaa) && (ptrInfo->FSI_TrailSig[2] == 0x55) &&  // check that the FSInfo sector is valid
                  (ptrInfo->FSI_LeadSig[3] == 0x41) && (ptrInfo->FSI_LeadSig[2] == 0x61) && (ptrInfo->FSI_LeadSig[1] == 0x52) && (ptrInfo->FSI_LeadSig[0] == 0x52) &&
                  (ptrInfo->FSI_StrucSig[3] == 0x61) && (ptrInfo->FSI_StrucSig[2] == 0x41) && (ptrInfo->FSI_StrucSig[1] == 0x72) && (ptrInfo->FSI_StrucSig[0] == 0x72)) {
                    utDisks[0].utFileInfo.ulFreeClusterCount = ((ptrInfo->FSI_Free_Count[3] << 24) + (ptrInfo->FSI_Free_Count[2] << 16) + (ptrInfo->FSI_Free_Count[1] << 8) + ptrInfo->FSI_Free_Count[0]);
                    utDisks[0].utFileInfo.ulNextFreeCluster = ((ptrInfo->FSI_Nxt_Free[3] << 24) + (ptrInfo->FSI_Nxt_Free[2] << 16) + (ptrInfo->FSI_Nxt_Free[1] << 8) + ptrInfo->FSI_Nxt_Free[0]);
                    if (utDisks[0].utFileInfo.ulNextFreeCluster <= (utDisks[0].utFileInfo.ulFreeClusterCount - 2)) { // check that the information is valid
                        utDisks[0].ucDiskFlags |= FSINFO_VALID;
                    }
                }
            }
            utDisks[0].ucDiskFlags |= DISK_MOUNTED;                      // the disk is now ready for use
            fnMemoryDebugMsg("Disk D mounted\r\n");
            iMemoryOperation |= _INITIALISED_MEMORY;
            iMemoryState = SD_STATE_READY;
            break;
#if defined UTFAT_WRITE && defined UTFAT_FORMATTING
        case SD_STATE_FORMATTING_DISK_1:                                 // start by creating a partition
            {
                EXTENDED_BOOT_RECORD *ptrExtendedBootRecord = (EXTENDED_BOOT_RECORD *)utDisks[0].ptrSectorData;
                unsigned long ulPartitionSize = (utDisks[0].ulSD_sectors - BOOT_SECTOR_LOCATION);
                uMemset(ptrExtendedBootRecord, 0, 512);                  // ensure content is blank
                ptrExtendedBootRecord->EBR_partition_table[0].starting_cylinder = 2;// fixed values
                ptrExtendedBootRecord->EBR_partition_table[0].starting_head     = 0x0c;
                ptrExtendedBootRecord->EBR_partition_table[0].partition_type    = 0x0b;
                ptrExtendedBootRecord->EBR_partition_table[0].ending_cylinder   = 0x38;
                ptrExtendedBootRecord->EBR_partition_table[0].ending_head       = 0xf8;
                ptrExtendedBootRecord->EBR_partition_table[0].ending_sector     = 0xb8;
                ptrExtendedBootRecord->EBR_partition_table[0].start_sector[0]   = BOOT_SECTOR_LOCATION;
                ptrExtendedBootRecord->EBR_partition_table[0].partition_size[0] = (unsigned char)(ulPartitionSize);
                ptrExtendedBootRecord->EBR_partition_table[0].partition_size[1] = (unsigned char)(ulPartitionSize >> 8);
                ptrExtendedBootRecord->EBR_partition_table[0].partition_size[2] = (unsigned char)(ulPartitionSize >> 16);
                ptrExtendedBootRecord->EBR_partition_table[0].partition_size[3] = (unsigned char)(ulPartitionSize >> 24);
                ptrExtendedBootRecord->ucCheck55 = 0x55;
                ptrExtendedBootRecord->ucCheckAA = 0xaa;
                if ((iActionResult = utCommitSector(&utDisks[0], 0)) != 0) {
                    if (iActionResult == CARD_BUSY_WAIT) {
                        return;                                          // write is taking time to complete so quit for the moment
                    }
                    iMemoryState = SD_NOT_FORMATTED;
                    break;
                }
    #ifdef _WINDOWS
                iFormatCount = 0;
    #endif
                iMemoryState = SD_STATE_FORMATTING_DISK_2;               // fall through
            }
        case SD_STATE_FORMATTING_DISK_2:                                 // add the boot sector
            {
                BOOT_SECTOR_FAT32 *ptrBootSector = (BOOT_SECTOR_FAT32 *)utDisks[0].ptrSectorData;
                unsigned long ulPartitionSize = (utDisks[0].ulSD_sectors - BOOT_SECTOR_LOCATION);
                uMemset(ptrBootSector, 0, 508);                          // ensure content is blank (without overwriting 0xaa550000)
                ptrBootSector->boot_sector_bpb.BS_jmpBoot[0]     = 0xeb; // fixed values
                ptrBootSector->boot_sector_bpb.BS_jmpBoot[1]     = 0x58;
                ptrBootSector->boot_sector_bpb.BS_jmpBoot[2]     = 0x90;
                uStrcpy(ptrBootSector->boot_sector_bpb.BS_OEMName, "MSDOS5.0");
                utDisks[0].utFAT.usBytesPerSector                = BYTES_PER_SECTOR;
                ptrBootSector->boot_sector_bpb.BPB_BytesPerSec[0]= (unsigned char)(BYTES_PER_SECTOR);
                ptrBootSector->boot_sector_bpb.BPB_BytesPerSec[1]= (unsigned char)(BYTES_PER_SECTOR >> 8);
                ptrBootSector->boot_sector_bpb.BPB_SecPerTrk[0]  = 63;
                ptrBootSector->boot_sector_bpb.BPB_NumHeads[0]   = 0xff;
                ptrBootSector->boot_sector_bpb.BPB_RsvdSecCnt[0] = RESERVED_SECTION_COUNT;
                ptrBootSector->boot_sector_bpb.BPB_NumFATs       = NUMBER_OF_FATS;
                ptrBootSector->boot_sector_bpb.BPB_Media         = FIXED_MEDIA;
                ptrBootSector->boot_sector_bpb.BPB_HiddSec[0]    = BOOT_SECTOR_LOCATION;
                ptrBootSector->boot_sector_bpb.BPB_TotSec32[0]   = (unsigned char)(ulPartitionSize);
                ptrBootSector->boot_sector_bpb.BPB_TotSec32[1]   = (unsigned char)(ulPartitionSize >> 8);
                ptrBootSector->boot_sector_bpb.BPB_TotSec32[2]   = (unsigned char)(ulPartitionSize >> 16);
                ptrBootSector->boot_sector_bpb.BPB_TotSec32[3]   = (unsigned char)(ulPartitionSize >> 24);
                if (ulPartitionSize <= 532480) {                         // disks up to 260MB
                    utDisks[0].utFAT.ucSectorsPerCluster = 1;
                }
                else if (ulPartitionSize <= 16777216) {                  // disks up to 8GB
                    utDisks[0].utFAT.ucSectorsPerCluster = 8;
                }
                else if (ulPartitionSize <= 33554432) {                  // disks up to 16GB
                    utDisks[0].utFAT.ucSectorsPerCluster = 16;
                }
                else if (ulPartitionSize <= 67108864) {                  // disks up to 32GB
                    utDisks[0].utFAT.ucSectorsPerCluster = 32;
                }
                else {                                                   // greater than 32GB
                    utDisks[0].utFAT.ucSectorsPerCluster = 64;
                }
                ptrBootSector->boot_sector_bpb.BPB_SecPerClus = utDisks[0].utFAT.ucSectorsPerCluster;
                ulFAT32size = (((256 * utDisks[0].utFAT.ucSectorsPerCluster) + NUMBER_OF_FATS)/2);
                ulFAT32size = (((ulPartitionSize - RESERVED_SECTION_COUNT) + (ulFAT32size - 1)) / ulFAT32size);
                ptrBootSector->BPB_FATSz32[0]   = (unsigned char)(ulFAT32size);
                ptrBootSector->BPB_FATSz32[1]   = (unsigned char)(ulFAT32size >> 8);
                ptrBootSector->BPB_FATSz32[2]   = (unsigned char)(ulFAT32size >> 16);
                ptrBootSector->BPB_FATSz32[3]   = (unsigned char)(ulFAT32size >> 24);
                ptrBootSector->BPB_RootClus[0]  = 2;
                ptrBootSector->BPB_FSInfo[0]    = 1;
                ptrBootSector->BPB_BkBootSec[0] = BACKUP_ROOT_SECTOR;
                ptrBootSector->BS_DrvNum        = 0x80;
                ptrBootSector->BS_BootSig       = 0x29;
    #ifdef RANDOM_NUMBER_GENERATOR
                ptrBootSector->BS_VolID[0] = (unsigned char)fnRandom();  // can also be generated by combining present data and time
                ptrBootSector->BS_VolID[1] = (unsigned char)fnRandom();
                ptrBootSector->BS_VolID[2] = (unsigned char)fnRandom();
                ptrBootSector->BS_VolID[3] = (unsigned char)fnRandom();
    #else
                ptrBootSector->BS_VolID[0] = 1;
                ptrBootSector->BS_VolID[1] = 2;
                ptrBootSector->BS_VolID[2] = 3;
                ptrBootSector->BS_VolID[3] = 4;
    #endif
                utDisks[0].ulLogicalBaseAddress = (BOOT_SECTOR_LOCATION + RESERVED_SECTION_COUNT + (ulFAT32size * NUMBER_OF_FATS));
                utDisks[0].ulVirtualBaseAddress = utDisks[0].ulLogicalBaseAddress - (2 * utDisks[0].utFAT.ucSectorsPerCluster);
                if (utDisks[0].cVolumeLabel[0] == 0) {
                    uMemcpy(ptrBootSector->BS_VolLab, "NO NAME    ", 11);
                }
                else {
                    int i = 0;
                    while (i < 11) {
                        if ((ptrBootSector->BS_VolLab[i] = utDisks[0].cVolumeLabel[i]) == 0) {
                            while (i < 11) {
                                ptrBootSector->BS_VolLab[i++] = ' ';
                            }
                        }
                        i++;
                    }
                }
                uMemcpy(ptrBootSector->BS_FilSysType, "FAT32   ", 8);
              //ptrBootSector->ucCheck55 = 0x55;                         // already exists in buffer
              //ptrBootSector->ucCheckAA = 0xaa;
                iMemoryState = SD_STATE_FORMATTING_DISK_3;               // fall through
            }
        case SD_STATE_FORMATTING_DISK_3:
            if ((iActionResult = utCommitSector(&utDisks[0], BOOT_SECTOR_LOCATION)) != 0) {
                if (iActionResult == CARD_BUSY_WAIT) {
                    return;                                              // write is taking time to complete so quit for the moment
                }
                iMemoryState = SD_NOT_FORMATTED;
                break;
            }
            iMemoryState = SD_STATE_FORMATTING_DISK_4;                   // fall through
        case SD_STATE_FORMATTING_DISK_4:
            if ((iActionResult = utCommitSector(&utDisks[0], (BOOT_SECTOR_LOCATION + BACKUP_ROOT_SECTOR))) != 0) {
                if (iActionResult == CARD_BUSY_WAIT) {
                    return;                                              // write is taking time to complete so quit for the moment
                }
                iMemoryState = SD_NOT_FORMATTED;
                break;
            }
            else {
                DIR_ENTRY_STRUCTURE_FAT32 *ptrVolumeEntry = (DIR_ENTRY_STRUCTURE_FAT32 *)utDisks[0].ptrSectorData;
                uMemset(utDisks[0].ptrSectorData, 0, 512);
                uMemcpy(ptrVolumeEntry->DIR_Name, utDisks[0].cVolumeLabel, 11); //add the volume ID
                ptrVolumeEntry->DIR_Attr = DIR_ATTR_VOLUME_ID;
                ulFatSectors = 0;
                iMemoryState = SD_STATE_FORMATTING_DISK_5;
            }
        case SD_STATE_FORMATTING_DISK_5:                                 // ensure that the first boot cluster is blank (with volume ID)
            if (ulFatSectors < utDisks[0].utFAT.ucSectorsPerCluster) {
                if ((iActionResult = utCommitSector(&utDisks[0], (ulFatSectors + utDisks[0].ulLogicalBaseAddress))) != 0) {
                    if (iActionResult == CARD_BUSY_WAIT) {
                        return;                                          // write is taking time to complete so quit for the moment
                    }
                    iMemoryState = SD_NOT_FORMATTED;
                    break;
                }
                ulFatSectors++;
                uMemset(utDisks[0].ptrSectorData, 0, 512);
                uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);          // schedule again to continue with next sector
                break;
            }
            iMemoryState = SD_STATE_FORMATTING_DISK_6;
            ulFatSectors = 0;
            ucFatCopies = 1;
        case SD_STATE_FORMATTING_DISK_6:
            uMemset(utDisks[0].ptrSectorData, 0, 512);
            uMemcpy(utDisks[0].ptrSectorData, ucEmptyFAT32, 12);
            iMemoryState = SD_STATE_FORMATTING_DISK_7;                   // fall through
        case SD_STATE_FORMATTING_DISK_7:
            if ((iActionResult = utCommitSector(&utDisks[0], (ulFatSectors + (BOOT_SECTOR_LOCATION + RESERVED_SECTION_COUNT)))) != 0) {
                if (iActionResult == CARD_BUSY_WAIT) {
                    return;                                              // write is taking time to complete so quit for the moment
                }
                iMemoryState = SD_NOT_FORMATTED;
                break;
            }
            uMemset(utDisks[0].ptrSectorData, 0, 512);
            ulFatSectors++;
            iMemoryState = SD_STATE_FORMATTING_DISK_8;                   // fall through
        case SD_STATE_FORMATTING_DISK_8:
            if (ulFatSectors >= (ulFAT32size * ucFatCopies)) {
                if (ucFatCopies < NUMBER_OF_FATS) {
                    ucFatCopies++;
                    iMemoryState = SD_STATE_FORMATTING_DISK_6;           // next FAT copy
                    uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);      // schedule again to continue with next sector
                    break;
                }
                iMemoryState = SD_STATE_FORMATTING_DISK_9;               // fall through since all FAT copies have been initialised
            }
            else {
                if ((iActionResult = utCommitSector(&utDisks[0], (ulFatSectors + (BOOT_SECTOR_LOCATION + RESERVED_SECTION_COUNT)))) != 0) {
                    if (iActionResult == CARD_BUSY_WAIT) {
                        return;                                          // write is taking time to complete so quit for the moment
                    }
                    iMemoryState = SD_NOT_FORMATTED;
                    break;
                }
                ulFatSectors++;
                if ((ulFatSectors % 256) == 0) {
                    fnMemoryDebugMsg("*");
                }
    #ifdef _WINDOWS
                if (++iFormatCount == 0x100) {                           // accelerate simulation
                    ulFatSectors = (ulFAT32size * ucFatCopies);
                    iFormatCount = 0;
                }
    #endif
                uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);          // schedule again to continue with next sector
                break;
            }
        case SD_STATE_FORMATTING_DISK_9:                                 // finally create the FAT32 info record
            {
                INFO_SECTOR_FAT32 *ptrInfoSector = (INFO_SECTOR_FAT32 *)utDisks[0].ptrSectorData;
              //unsigned long ulFreeCount = ((ulFAT32size * (512/sizeof(signed long))) - 3); // total free clusters minus default root cluster and 2 unused ones
                unsigned long ulFreeCount = ((utDisks[0].ulSD_sectors - RESERVED_SECTION_COUNT - BOOT_SECTOR_LOCATION - (NUMBER_OF_FATS * ulFAT32size))/utDisks[0].utFAT.ucSectorsPerCluster); // {1}
                ptrInfoSector->FSI_LeadSig[3] = 0x41;
                ptrInfoSector->FSI_LeadSig[2] = 0x61;
                ptrInfoSector->FSI_LeadSig[1] = 0x52;
                ptrInfoSector->FSI_LeadSig[0] = 0x52;
                ptrInfoSector->FSI_StrucSig[3] = 0x61;
                ptrInfoSector->FSI_StrucSig[2] = 0x41;
                ptrInfoSector->FSI_StrucSig[1] = 0x72;
                ptrInfoSector->FSI_StrucSig[0] = 0x72;                                 
                fnAddInfoSect(ptrInfoSector, (ulFreeCount - 1), 3);      // one cluster occupied by root directory by default and first useable cluster number is 3
                ptrInfoSector->FSI_StrucSig[3] = 0x61;
                ptrInfoSector->FSI_StrucSig[2] = 0x41;
                ptrInfoSector->FSI_StrucSig[1] = 0x72;
                ptrInfoSector->FSI_StrucSig[0] = 0x72;
                ptrInfoSector->FSI_TrailSig[3] = 0xaa;
                ptrInfoSector->FSI_TrailSig[2] = 0x55;
                if ((iActionResult = utCommitSector(&utDisks[0], (BOOT_SECTOR_LOCATION + 1))) != 0) { // write info sector
                    if (iActionResult == CARD_BUSY_WAIT) {
                        return;                                          // write is taking time to complete so quit for the moment
                    }
                    iMemoryState = SD_NOT_FORMATTED;
                    break;
                }
                iMemoryState = SD_STATE_FORMATTING_DISK_10;
                utDisks[0].ucDiskFlags |= FSINFO_VALID;
        case SD_STATE_FORMATTING_DISK_10:                                // backup
                if ((iActionResult = utCommitSector(&utDisks[0], (BOOT_SECTOR_LOCATION + 1 + BACKUP_ROOT_SECTOR))) != 0) { // write info sector backup
                    if (iActionResult == CARD_BUSY_WAIT) {
                        return;                                          // write is taking time to complete so quit for the moment
                    }
                    iMemoryState = SD_NOT_FORMATTED;
                    break;
                }
                fnMemoryDebugMsg("Disk D formatted\r\n");
                iMemoryState = SD_MOUNTING_1;
                uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);          // schedule again to mount the disk which has just been formatted
            }
            break;
#endif
        }
    }
    else if (iMemoryOperation & _COUNTING_CLUSTERS) {                    // cluster counting in progress
        if (ulClusterSectorCheck >= utDisks[0].utFAT.ulFAT_start) {
            int iRepetitions = 10;
            unsigned long ulSectorContent[512/sizeof(signed long)];
            if (ulClusterSectorCheck < (iRepetitions + utDisks[0].utFAT.ulFAT_start)) {
                iRepetitions = (ulClusterSectorCheck - utDisks[0].utFAT.ulFAT_start);
            }
            while (iRepetitions-- >= 0) {
                if ((utReadDiskData(&utDisks[0], ulClusterSectorCheck--, ulSectorContent, 0, 512)) == UTFAT_SUCCESS) { // read a FAT sector containing the cluster information
                    int i = 0;
                    while (i < (512/sizeof(signed long))) {
                        if (ulSectorContent[i++] == 0) {
                            ulActiveFreeClusterCount++;
                        }
                    }
                }
            }
            uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);              // schedule to run again
        }
        else {
            iMemoryOperation &= ~_COUNTING_CLUSTERS;
            utDisks[0].utFileInfo.ulFreeClusterCount = ulActiveFreeClusterCount;
            fnInterruptMessage(cluster_task, UTFAT_OPERATION_COMPLETED); // inform that the cluster count has completed
            cluster_task = 0;
        }
    }

    while (fnRead(ptrTaskTable->TaskID, ucInputMessage, HEADER_LENGTH)) {// check task input queue
        switch (ucInputMessage[MSG_SOURCE_TASK]) {
        case TIMER_EVENT:
        case INTERRUPT_EVENT:
            break;
        }
    }
}


static void fnCardNotFormatted(void)
{
    utDisks[0].ucDiskFlags = DISK_UNFORMATTED;
    iMemoryState = SD_NOT_FORMATTED;
    fnMemoryDebugMsg("None-formatted SD-card\r\n");
}

// Interrupt initisalisation sequence on error, remove power and start next try after a delay
//
static void fnInitialisationError(void)
{
    SET_SD_CS_HIGH();
    POWER_DOWN_SD_CARD();
    iMemoryState = SD_STATE_WAIT_CARD;                                   // wait for next check of valid card
    uTaskerMonoTimer(OWN_TASK, T_NEXT_CHECK, E_POLL_SD_CARD);            // try again later
    fnMemoryDebugMsg("No valid SD-card\r\n");
}


static const unsigned char *fnCreateCommand(unsigned char ucCommand, unsigned long ulValue)
{
    static unsigned char ucCommandBuffer[6];                             // space for a command
    ucCommandBuffer[0] = ucCommand;
    ucCommandBuffer[1] = (unsigned char)(ulValue >> 24);
    ucCommandBuffer[2] = (unsigned char)(ulValue >> 16);
    ucCommandBuffer[3] = (unsigned char)(ulValue >> 8);
    ucCommandBuffer[4] = (unsigned char)(ulValue);
    ucCommandBuffer[5] = 0;
    return (const unsigned char *)ucCommandBuffer;
}



// Read a specified amount of data from SD card into a data buffer
//
static int fnGetSector(unsigned char *ptrBuf)
{
    unsigned char ucResponse;
    int iLength = 512;
    // this may block so a timer may be required to protect it (100ms?)
    do {
        WRITE_SPI_CMD(0xff);
        WAIT_TRANSMISSON_END();                                          // wait until transmission complete
    } while ((ucResponse = READ_SPI_DATA()) == 0xff);
    if (ucResponse != 0xfe) {
        return UTFAT_DISK_READ_ERROR;                                    // error
    }
    do {
        WRITE_SPI_CMD(0xff);
        WAIT_TRANSMISSON_END();                                          // wait until transmission complete
        *ptrBuf++ = READ_SPI_DATA();
    } while (--iLength);
    WRITE_SPI_CMD(0xff);                                                 // read and discard two CRC bytes
    WAIT_TRANSMISSON_END();                                              // wait until transmission complete
    ucResponse = READ_SPI_DATA();
    WRITE_SPI_CMD(0xff);
    WAIT_TRANSMISSON_END();                                              // wait until transmission complete
    ucResponse = READ_SPI_DATA();
    return UTFAT_SUCCESS;                                                // read successfully terminated
}

// Write a specified amount of data from data buffer to SD card
//
static int fnPutSector(unsigned char *ptrBuf)
{
    #define DATA_TOKEN 0xfe
    unsigned char ucResponse;
    int iLength = 512;
    // this may block so a timer may be required to protect it (100ms?)
    do {
        WRITE_SPI_CMD(0xff);
        WAIT_TRANSMISSON_END();                                          // wait until transmission complete
    } while ((ucResponse = READ_SPI_DATA()) != 0xff);
    WRITE_SPI_CMD(DATA_TOKEN);                                           // transmit data byte
    WAIT_TRANSMISSON_END();                                              // wait until transmission complete
    READ_SPI_DATA();                                                     // clear the receiver
    do {
        WRITE_SPI_CMD(*ptrBuf);                                          // transmit data byte
        ptrBuf++;
        WAIT_TRANSMISSON_END();                                          // wait until transmission complete
        READ_SPI_DATA();                                                 // clear the receiver
    } while (--iLength);
    WRITE_SPI_CMD(0xff);                                                 // send two dummy CRC bytes
    WAIT_TRANSMISSON_END();                                              // wait until transmission complete
    READ_SPI_DATA();
    WRITE_SPI_CMD(0xff);
    WAIT_TRANSMISSON_END();                                              // wait until transmission complete
    READ_SPI_DATA();
    WRITE_SPI_CMD(0xff);                                                 // send two dummy CRC bytes
    WAIT_TRANSMISSON_END();                                              // wait until transmission complete
    ucResponse = READ_SPI_DATA();
    if ((ucResponse & 0x1f) != 0x05) {
        return UTFAT_DISK_WRITE_ERROR;
    }
    return UTFAT_SUCCESS;                                                // write successfully completed
}

// Write a specified amount of data from data buffer to SD card
//
static int fnSetSector(unsigned char ucValue)
{
    #define DATA_TOKEN 0xfe
    unsigned char ucResponse;
    int iLength = 512;
    // this may block so a timer may be required to protect it (100ms?)
    do {
        WRITE_SPI_CMD(0xff);
        WAIT_TRANSMISSON_END();                                          // wait until transmission complete
    } while ((ucResponse = READ_SPI_DATA()) != 0xff);
    WRITE_SPI_CMD(DATA_TOKEN);                                           // transmit data byte
    WAIT_TRANSMISSON_END();                                              // wait until transmission complete
    READ_SPI_DATA();                                                     // clear the receiver
    do {
        WRITE_SPI_CMD(ucValue);                                          // transmit data byte
        WAIT_TRANSMISSON_END();                                          // wait until transmission complete
        READ_SPI_DATA();                                                 // clear the receiver
    } while (--iLength);
    WRITE_SPI_CMD(0xff);                                                 // send two dummy CRC bytes
    WAIT_TRANSMISSON_END();                                              // wait until transmission complete
    READ_SPI_DATA();
    WRITE_SPI_CMD(0xff);
    WAIT_TRANSMISSON_END();                                              // wait until transmission complete
    READ_SPI_DATA();
    WRITE_SPI_CMD(0xff);                                                 // send two dummy CRC bytes
    WAIT_TRANSMISSON_END();                                              // wait until transmission complete
    ucResponse = READ_SPI_DATA();
    if ((ucResponse & 0x1f) != 0x05) {
        return UTFAT_DISK_WRITE_ERROR;
    }
    return UTFAT_SUCCESS;                                                // write successfully completed
}


// Read a specified amount of data from SD card into a data buffer
//
static int fnReadPartialSector(unsigned char *ptrBuf, unsigned short usStart, unsigned short usStop)
{
    unsigned char ucResponse;
    unsigned short usLoopCounter = 0;
    // this may block so a timer may be required to protect it (100ms?)
    do {
        WRITE_SPI_CMD(0xff);
        WAIT_TRANSMISSON_END();                                          // wait until transmission complete
    } while ((ucResponse = READ_SPI_DATA()) == 0xff);
    if (ucResponse != 0xfe) {
        fnMemoryDebugMsg("ERR-1!!!\r\n");
        return 0;                                                        // error
    }
    while (usLoopCounter < usStart) {
        WRITE_SPI_CMD(0xff);
        usLoopCounter++;
        WAIT_TRANSMISSON_END();                                          // wait until transmission complete
        READ_SPI_DATA();                                                 // discard unwanted data at the start of the sector
    }
    while (usLoopCounter < usStop) {
        WRITE_SPI_CMD(0xff);
        usLoopCounter++;
        WAIT_TRANSMISSON_END();                                          // wait until transmission complete
        *ptrBuf++ = READ_SPI_DATA();
    }
    while (usLoopCounter < 512) {
        WRITE_SPI_CMD(0xff);
        usLoopCounter++;
        WAIT_TRANSMISSON_END();                                          // wait until transmission complete
        READ_SPI_DATA();                                                 // discard any additional data, but always read a full 512 byte block
    }
    WRITE_SPI_CMD(0xff);                                                 // read and discard two CRC bytes
    WAIT_TRANSMISSON_END();                                              // wait until transmission complete
    ucResponse = READ_SPI_DATA();
    WRITE_SPI_CMD(0xff);
    WAIT_TRANSMISSON_END();                                              // wait until transmission complete
    ucResponse = READ_SPI_DATA();
    return 1;                                                            // read terminated
}


// Send a command to the SD-card and return a result and optionally 4 additional returned arguments
//
static int fnSendSD_command(const unsigned char ucCommand[6], unsigned char *ucResult, unsigned char *ptrReturnData)
{
    static int iCommandYieldCount = 0;
    static int iCommandState = 0;
    int iRtn = 0;

    if (iCommandYieldCount > CommandTimeout) {                           // the present access is taking too long - quit with SD card error
        fnMemoryDebugMsg("TIMEOUT!!!\r\n");
        iRtn = ERROR_CARD_TIMEOUT;
    }
    else {
        switch (iCommandState) {
        case 0:
            if (fnWaitSD_ready(20) != 0) {                               // poll up to 20 times before yielding
                iCommandYieldCount++;                                    // monitor the maximum delay
                uTaskerStateChange(OWN_TASK, UTASKER_GO);                // switch to polling mode of operation
              //fnMemoryDebugMsg("B");
                return CARD_BUSY_WAIT;
            }                                                            // wait until the SD card is ready
#ifdef SD_CONTROLLER_AVAILABLE
            MCIArgument = (ucCommand[1] | (ucCommand[2] << 8) | (ucCommand[3] << 16) | (ucCommand[4] << 24)); // set the argument of the command
            MCICommand = (ucCommand[0] | MCICmd_Enable);                 // launch the command
#else
            for ( ; iCommandState < 6; iCommandState++) {                // command content is always 6 bytes in length
                WRITE_SPI_CMD(ucCommand[iCommandState]);                 // send the command and arguments
                WAIT_TRANSMISSON_END();                                  // wait until transmission complete
                READ_SPI_DATA();                                         // read to clear the flag
            }                                                            // fall through intentional after sending the command
#endif
            iCommandYieldCount = 0;
        case 6:
            do {
                if (iCommandState++ > 26) {
                    iCommandYieldCount++;                                // monitor the maximum delay
                    iCommandState = 6;
                    uTaskerStateChange(OWN_TASK, UTASKER_GO);            // switch to polling mode of operation
                  //fnMemoryDebugMsg("b");
                    return CARD_BUSY_WAIT;                               // poll up to 20 times before yielding
                }
#ifdef SD_CONTROLLER_AVAILABLE
                if (MCIStatus & CmdTimeOut) {                            // timeout occurred while waiting for command to complete
                    MCIClear = (CmdCrcFailClr | DataCrcFailClr | CmdTimeOutClr | DataTimeOutClr | TxUnderrunClr | RxOverrunClr | CmdRespEndClr | CmdSentClr | DataEndClr | StartBitErrClr | DataBlockEndClr);
                    MCICommand = 0;
                    MCIArgument = 0xffffffff;
                }
                else if (MCIStatus & CmdCrcFail) {                       // CRC error detected
                }
                else if (MCIStatus & CmdRespEnd) {                       // command has successfully completed
    #ifdef _WINDOWS
                    if (ucCommand[0] == SEND_OP_COND_ACMD_CMD41) {
                        MCIResponse0 = 0;
                    }
                    else {
                        MCIResponse0 = R1_IN_IDLE_STATE;
                    }
    #endif
                    MCIClear = (CmdCrcFailClr | DataCrcFailClr | CmdTimeOutClr | DataTimeOutClr | TxUnderrunClr | RxOverrunClr | CmdRespEndClr | CmdSentClr | DataEndClr | StartBitErrClr | DataBlockEndClr);
                  //*ucResult = (unsigned char)MCIRespCmd;               // the response received
                    *ucResult = (unsigned char)MCIResponse0;             // the result received
                    break;
                }
                else {
                    *ucResult = SD_CARD_BUSY;
    #ifdef _WINDOWS
                    MCIStatus = CmdRespEnd;
    #endif
                }
#else
                WRITE_SPI_CMD(0xff);                                     // send idle line
                WAIT_TRANSMISSON_END();                                  // wait until transmission complete
                *ucResult = READ_SPI_DATA();                             // read result byte
#endif
            } while (*ucResult & SD_CARD_BUSY);                          // poll the card until it is no longer indicating busy and returns the value
            if (ptrReturnData != 0) {                                    // if the caller requests data, read it here
                int iReturnLength;
                if (ucCommand[0] == SEND_CSD_CMD9) {                     // exception requiring 18 bytes
                    iReturnLength = 18;
                }
                else {
                    iReturnLength = 4;
                }
#ifdef SD_CONTROLLER_AVAILABLE
    #ifdef _WINDOWS
                MCIResponse2 = VOLTAGE_2_7__3_6;
                MCIResponse3 = CHECK_PATTERN;
    #endif
                *ptrReturnData++ = MCIResponse0;
                *ptrReturnData++ = MCIResponse1;
                *ptrReturnData++ = MCIResponse2;
                *ptrReturnData   = MCIResponse3;
#else
                for (iCommandState = 0; iCommandState < iReturnLength; iCommandState++) {
                    WRITE_SPI_CMD(0xff);                                 // send the command and arguments
                    WAIT_TRANSMISSON_END();                              // wait until transmission complete
                    *ptrReturnData++ = READ_SPI_DATA();
                }
#endif
            }
            break;
        }
    }
    if (iCommandYieldCount != 0) {
        uTaskerStateChange(OWN_TASK, UTASKER_STOP);                      // switch to event mode of operation since write has completed
    }
    iCommandState = 0;
    iCommandYieldCount = 0;
    return iRtn;
}

// Wait until the SD card is ready by reading until 0xff is returned
//
static int fnWaitSD_ready(int iMaxWait)
{
#ifdef SD_CONTROLLER_AVAILABLE
    while (MCIStatus & CmdActive) {                                      // clear any command activity
        MCICommand = 0;
        MCIClear = (CmdCrcFailClr | DataCrcFailClr | CmdTimeOutClr | DataTimeOutClr | TxUnderrunClr | RxOverrunClr | CmdRespEndClr | CmdSentClr | DataEndClr | StartBitErrClr | DataBlockEndClr);
    #ifdef _WINDOWS
        MCIStatus = 0;
    #endif
    }
    // delay??
#else
    do {
        if (iMaxWait-- == 0) {
            return 1;                                                    // maximum wait attempts executed
        }
        WRITE_SPI_CMD(0xff);
        WAIT_TRANSMISSON_END();                                          // wait until transmission complete
    } while (READ_SPI_DATA() != 0xff);                                   // 0xff expected when the device is ready
#endif
    return 0;                                                            // the device is now ready
}


static int utReadDisk(UTDISK *ptr_utDisk, unsigned long ulSectorNumber, unsigned char ucSectorCount)
{
    static unsigned long ulSector;
    static int iActionResult;
    switch (iMemoryOperation & _READING_MEMORY) {
    case _IDLE_MEMORY:
        if (!(ptr_utDisk->ucDiskFlags & HIGH_CAPACITY_SD_CARD)) {
            ulSectorNumber *= 512;                                       // convert the sector number to byte address
        }
        SET_SD_CS_LOW();
        iMemoryOperation |= _READING_MEMORY;
        ulSector = ulSectorNumber;
    case _READING_MEMORY:
        if (ucSectorCount == 1) {
            unsigned char ucResult;
            while ((iActionResult = fnSendSD_command(fnCreateCommand(READ_SINGLE_BLOCK_CMD17, ulSector), &ucResult, 0)) == CARD_BUSY_WAIT) {}
            if (iActionResult < 0) {
                SET_SD_CS_HIGH();
                iMemoryOperation &= ~_READING_MEMORY;                    // read operation has failed
                return iActionResult;
            }
            if (ucResult == 0) {
                fnGetSector(ptr_utDisk->ptrSectorData);                  // start reading single sector to buffer
                if (!(ptr_utDisk->ucDiskFlags & HIGH_CAPACITY_SD_CARD)) {
                    ulSector /= 512;                                     // convert back to sector number from byte offset
                }
                ptr_utDisk->ulPresentSector = ulSector;                  // update the sector which is presently valid
            }
            SET_SD_CS_HIGH();
            iMemoryOperation &= ~_READING_MEMORY;                        // read operation has completed
        }
        else {
        }
        break;
    }
    return UTFAT_SUCCESS;
}


static int utReadDiskData(UTDISK *ptr_utDisk, unsigned long ulSectorNumber, void *ptrBuf, unsigned short usOffset, unsigned short usLength)
{
    static unsigned long ulSector;
    static int iActionResult;
    switch (iMemoryOperation & _READING_MEMORY) {
    case _IDLE_MEMORY:
        if (!(ptr_utDisk->ucDiskFlags & HIGH_CAPACITY_SD_CARD)) {
            ulSectorNumber *= 512;                                       // convert the sector number to byte address
        }
        SET_SD_CS_LOW();
        iMemoryOperation |= _READING_MEMORY;
        ulSector = ulSectorNumber;
    case _READING_MEMORY:
        if (/*ucSectorCount ==*/ 1) {
            unsigned char ucResult;
            while ((iActionResult = fnSendSD_command(fnCreateCommand(READ_SINGLE_BLOCK_CMD17, ulSector), &ucResult, 0)) == CARD_BUSY_WAIT) {}
            if (iActionResult < 0) {
                SET_SD_CS_HIGH();
                iMemoryOperation &= ~_READING_MEMORY;                    // read operation has completed
                return iActionResult;
            }
            if (ucResult == 0) {
                fnReadPartialSector(ptrBuf, usOffset, (unsigned short)(usOffset + usLength));// start reading a sector direct to buffer
            }
            SET_SD_CS_HIGH();
            iMemoryOperation &= ~_READING_MEMORY;                        // read operation has completed
        }
        else {
        }
        break;
    }
    return UTFAT_SUCCESS;
}


// Read one or more sectors from the disk specified by the drive number
//
static int ut_read_disk(UTDISK *ptr_utDisk, unsigned char ucReadSize)
{
    if (utReadDisk(ptr_utDisk, ptr_utDisk->ulPresentSector, ucReadSize) != 0) { // read specified sector to buffer
        return CARD_BUSY_WAIT;
    }
    else {                                                               // read from SD media was successful and we expect to have the boot sector in the buffer
        BOOT_SECTOR_FAT32 *ptrBootSector = (BOOT_SECTOR_FAT32 *)ptr_utDisk->ptrSectorData;
        if ((ptrBootSector->ucCheck55 != 0x55) || (ptrBootSector->ucCheckAA != 0xaa)) {
            return ERROR_SECTOR_INVALID;                                 // the sector data is no valid
        }
    }
    return 0;
}


#ifdef UTFAT_LFN_READ
static int fnMatchLongName(const CHAR **ptrLongFileName, const CHAR *ptrMatchStart, LFN_ENTRY_STRUCTURE_FAT32 *ptrLFN_entry)
{
    int i = 13;                                                          // there are maximum 13 characters in an entry
    unsigned char ucCharacter = 0;
    unsigned char ucMatchCharacter;
    if ((ptrLFN_entry->LFN_Attribute & DIR_ATTR_MASK) != DIR_ATTR_LONG_NAME) {
        if (ptrLFN_entry->LFN_EntryNumber == 0) {                        // end of directory entries
            return END_DIRECTORY_ENTRIES;
        }
        return MATCH_CONTINUE;                                           // this is not a long file name entry so ignore it
    }
    while (--i >= 0) {
        switch (i) {
        case 12:
            ucCharacter = ptrLFN_entry->LFN_Name_12;
            break;
        case 11:
            ucCharacter = ptrLFN_entry->LFN_Name_11;
            break;
        case 10:
            ucCharacter = ptrLFN_entry->LFN_Name_10;
            break;
        case 9:
            ucCharacter = ptrLFN_entry->LFN_Name_9;
            break;
        case 8:
            ucCharacter = ptrLFN_entry->LFN_Name_8;
            break;
        case 7:
            ucCharacter = ptrLFN_entry->LFN_Name_7;
            break;
        case 6:
            ucCharacter = ptrLFN_entry->LFN_Name_6;
            break;
        case 5:
            ucCharacter = ptrLFN_entry->LFN_Name_5;
            break;
        case 4:
            ucCharacter = ptrLFN_entry->LFN_Name_4;
            break;
        case 3:
            ucCharacter = ptrLFN_entry->LFN_Name_3;
            break;
        case 2:
            ucCharacter = ptrLFN_entry->LFN_Name_2;
            break;
        case 1:
            ucCharacter = ptrLFN_entry->LFN_Name_1;
            break;
        case 0:
            ucCharacter = ptrLFN_entry->LFN_Name_0;
            break;
        }
        if (ucCharacter == 0xff) {                                       // past end of entry
            continue;
        }
        if (ucCharacter == 0x00) {                                       // null-terminator
            continue;
        }
        ucMatchCharacter = *(--(*ptrLongFileName));
        if (ucCharacter != ucMatchCharacter) {                           // check the match, working backward, with pre-decrement
            if ((ucCharacter >= 'A') && (ucCharacter <= 'Z')) {
                ucCharacter += ('a' - 'A');
            }
            else if ((ucCharacter >= 'a') && (ucCharacter <= 'z')) {
                ucMatchCharacter += ('a' - 'A');
            }
            else {
                return MATCH_FALSE;
            }
            if (ucCharacter != ucMatchCharacter) {                       // last chance for case-insensitive match
                return MATCH_FALSE;
            }
        }
    }
    if ((ptrLFN_entry->LFN_EntryNumber & ~0x40) == 1) {                  // the end of the long file name
        if (*ptrLongFileName == ptrMatchStart) {
            return MATCH_SUCCESSFUL;
        }
        return MATCH_FALSE;
    }
    return MATCH_CONTINUE;
}
#endif



// Load a sector if the present sector content is not alread valid
//
static int fnLoadSector(UTDISK *ptr_utDisk, unsigned long ulSector)
{
    if ((ulSector != 0) && (ptr_utDisk->ulPresentSector != ulSector)) {  // if the requested sector is not the already loaded one
#ifdef UTFAT_WRITE
        if (ptr_utDisk->ucDiskFlags & WRITEBACK_BUFFER_FLAG) {           // if changes have been made to the sector since its read it must be written back
            while (utCommitSector(ptr_utDisk, ptr_utDisk->ulPresentSector) == CARD_BUSY_WAIT) {} // force writeback to finalise the operation
            ptr_utDisk->ucDiskFlags &= ~WRITEBACK_BUFFER_FLAG;
        }
#endif
        ptr_utDisk->ulPresentSector = ulSector;
        return (utReadDisk(ptr_utDisk, ulSector, 1));                    // read the new sector
    }
    return UTFAT_SUCCESS;
}

// Read a sector directly to a buffer (length is maximum 512 bytes)
//
static int fnLoadData(UTDISK *ptr_utDisk, unsigned long ulSector, void *ptrBuf, unsigned short usOffset, unsigned short usLength)
{
    if ((ptr_utDisk->ulPresentSector == ulSector) && (!(ptr_utDisk->ucDiskFlags & WRITEBACK_BUFFER_FLAG))) { // if the requested sector is already loaded and hasn't been modified
        uMemcpy(ptrBuf, &ptr_utDisk->ptrSectorData[usOffset], (usLength - usOffset));
    }
    else {                                                               // the normal case is to load directly
        return (utReadDiskData(ptr_utDisk, ulSector, ptrBuf, usOffset, usLength)); // read data from disk without changing the present sector
    }
    return UTFAT_SUCCESS;
}



static int fnNextSector(UTDISK *ptr_utDisk, FILE_LOCATION *ptr_location)
{
    ptr_location->ulSector++;                                            // next sector
    if (((ptr_location->ulSector - ptr_utDisk->ulLogicalBaseAddress) % ptr_utDisk->utFAT.ucSectorsPerCluster) == 0) { // check whether the present cluster end has been reached
        unsigned long ulClusterSector = (ptr_utDisk->utFAT.ulFAT_start + (ptr_location->ulCluster >> 7)); // section where the FAT responsible for this cluster resides
        unsigned long ulCluster;
        if (fnLoadData(ptr_utDisk, ulClusterSector, (unsigned char *)&ulCluster, (unsigned short)((ptr_location->ulCluster & 0x7f) * sizeof(unsigned long)), sizeof(unsigned long)) != UTFAT_SUCCESS) { // read directly to cluster value
            return UTFAT_DISK_READ_ERROR;
        }
        ulCluster = LITTLE_LONG_WORD(ulCluster);                         // ensure endien is correct
        if ((ulCluster <= ptr_utDisk->ulDirectoryBase) || (ulCluster > ptr_utDisk->utFAT.ulClusterCount)) {
            return UTFAT_DIRECTORY_AREA_EXHAUSTED;
        }
        ptr_location->ulCluster = ulCluster;                             // set next cluster
        ptr_location->ulSector = ((ptr_location->ulCluster * ptr_utDisk->utFAT.ucSectorsPerCluster) + ptr_utDisk->ulVirtualBaseAddress);
    }
    return UTFAT_SUCCESS;
}


static int fnNextSectorCreate(UTDISK *ptr_utDisk, FILE_LOCATION *ptr_location)
{
    ptr_location->ulSector++;                                            // next sector
    if (((ptr_location->ulSector - ptr_utDisk->ulLogicalBaseAddress) % ptr_utDisk->utFAT.ucSectorsPerCluster) == 0) { // check whether the present clust end has been reached
        unsigned long ulClusterSector = (ptr_utDisk->utFAT.ulFAT_start + (ptr_location->ulCluster >> 7)); // section where the FAT responsible for this cluster resides
        unsigned long ulCluster;
        if (fnLoadData(ptr_utDisk, ulClusterSector, (unsigned char *)&ulCluster, (unsigned short)((ptr_location->ulCluster & 0x7f) * sizeof(unsigned long)), sizeof(unsigned long)) != UTFAT_SUCCESS) { // read directly to cluster value
            return UTFAT_DISK_READ_ERROR;
        }
        ulCluster = LITTLE_LONG_WORD(ulCluster);                         // ensure endien is correct
        if ((ulCluster <= ptr_utDisk->ulDirectoryBase) || (ulCluster >= CLUSTER_MASK)) { // invalid or end of current cluster - add a cluster so that the file can grow
            ulCluster = fnAllocateCluster(ptr_utDisk, ptr_location->ulCluster, NEW_RELATIVE_CLUSTER); // next cluster
        }
        if (ptr_location->ulCluster > ptr_utDisk->utFAT.ulClusterCount) {
            return UTFAT_DIRECTORY_AREA_EXHAUSTED;
        }
        ptr_location->ulCluster = ulCluster;
        ptr_location->ulSector = ((ulCluster * ptr_utDisk->utFAT.ucSectorsPerCluster) + ptr_utDisk->ulVirtualBaseAddress);
    }
    return UTFAT_SUCCESS;
}


static int fnDirectorySectorCreate(UTDISK *ptr_utDisk, FILE_LOCATION *ptr_location)
{
    unsigned long ulClusterSector = (ptr_utDisk->utFAT.ulFAT_start + (ptr_location->ulCluster >> 7)); // section where the FAT responsible for this cluster resides
    unsigned long ulCluster;
    if (fnLoadData(ptr_utDisk, ulClusterSector, (unsigned char *)&ulCluster, (unsigned short)((ptr_location->ulCluster & 0x7f) * sizeof(unsigned long)), sizeof(unsigned long)) != UTFAT_SUCCESS) { // read directly to cluster value
        return UTFAT_DISK_READ_ERROR;
    }
    ulCluster = LITTLE_LONG_WORD(ulCluster);                         // ensure endien is correct
    if ((ulCluster <= ptr_utDisk->ulDirectoryBase) || (ulCluster >= CLUSTER_MASK)) { // invalid or end of current cluster - add a cluster so that the file can grow
        ulCluster = fnAllocateCluster(ptr_utDisk, ptr_location->ulCluster, (INITIALISE_DIR_EXTENSION | NEW_RELATIVE_CLUSTER)); // next cluster
    }
    if (ptr_location->ulCluster > ptr_utDisk->utFAT.ulClusterCount) {
        return UTFAT_DIRECTORY_AREA_EXHAUSTED;
    }
    ptr_location->ulCluster = ulCluster;
    ptr_location->ulSector = ((ulCluster * ptr_utDisk->utFAT.ucSectorsPerCluster) + ptr_utDisk->ulVirtualBaseAddress);
    return UTFAT_SUCCESS;
}



static int fnNextDirectoryEntry(UTDISK *ptr_utDisk, DISK_LOCATION *ptr_location)
{
    if (++(ptr_location->ucDirectoryEntry) >= (ptr_utDisk->utFAT.usBytesPerSector/sizeof(DIR_ENTRY_STRUCTURE_FAT32))) { // end of present sector reached
        ptr_location->ucDirectoryEntry = 0;                              // start at first entry in next sector
        return (fnNextSector(ptr_utDisk, &ptr_location->directory_location));
    }
    return UTFAT_SUCCESS;
}

static void fnLoadShortFileInfo(UTFILEINFO *ptr_utFileInfo, const DIR_ENTRY_STRUCTURE_FAT32 *ptrDirectoryEntry)
{
    int i;
    unsigned char c;
    CHAR *ptrShortName = ptr_utFileInfo->cFileName;
    unsigned char ucNT_info = ptrDirectoryEntry->DIR_NTRes;

    for (i = 0; i < 8; i++) {                                            // name
        c = (unsigned char)ptrDirectoryEntry->DIR_Name[i];
        if (c == ' ') {
            break;
        }
        if (c == 0x05) {
            c = DIR_NAME_FREE;
        }
        if ((ucNT_info & 0x08) && (c >= 'A') && (c <= 'Z')) {
            c += ('a' - 'A');                                            // convert to small letters
        }
        *ptrShortName++ = c;
    }
    if (ptrDirectoryEntry->DIR_Name[8] != ' ') {                         // extension
        *ptrShortName++ = '.';
        for (i = 8; i < 11; i++) {
            c = ptrDirectoryEntry->DIR_Name[i];
            if (c == ' ') {
                break;
            }
            if ((ucNT_info & 0x10) && (c >= 'A') && (c <= 'Z')) {
                c += ('a' - 'A');                                        // convert to small letters
            }
            *ptrShortName++ = c;
        }
    }
    *ptrShortName = 0;                                                   // terminate

    ptr_utFileInfo->ucFileAttributes = ptrDirectoryEntry->DIR_Attr;
    ptr_utFileInfo->ulFileSize = ((ptrDirectoryEntry->DIR_FileSize[3] << 24) + (ptrDirectoryEntry->DIR_FileSize[2] << 16) + (ptrDirectoryEntry->DIR_FileSize[1] << 8) + ptrDirectoryEntry->DIR_FileSize[0]);
    ptr_utFileInfo->usFileData = ((ptrDirectoryEntry->DIR_WrtDate[1] << 8) + ptrDirectoryEntry->DIR_WrtDate[0]);
    ptr_utFileInfo->usFileTime = ((ptrDirectoryEntry->DIR_WrtTime[1] << 8) + ptrDirectoryEntry->DIR_WrtTime[0]);
}

#ifdef UTFAT_LFN_READ
static void fnExtractLongFileName(CHAR cLongFileName[100], LFN_ENTRY_STRUCTURE_FAT32 *ptrLFN_Entry, unsigned char *ucLongFileNameLength)
{
    int i = *ucLongFileNameLength;
    if (ptrLFN_Entry->LFN_EntryNumber & 0x40) {                          // last LFN entry
        if ((ptrLFN_Entry->LFN_Name_12 != 0xff) && (ptrLFN_Entry->LFN_Name_12 != 0x00)) { // string ending without null terminator
            cLongFileName[--i] = 0;                                      // force a null terminator
        }
    }
    cLongFileName[--i] = ptrLFN_Entry->LFN_Name_12;
    cLongFileName[--i] = ptrLFN_Entry->LFN_Name_11;
    cLongFileName[--i] = ptrLFN_Entry->LFN_Name_10;
    cLongFileName[--i] = ptrLFN_Entry->LFN_Name_9;
    cLongFileName[--i] = ptrLFN_Entry->LFN_Name_8;
    cLongFileName[--i] = ptrLFN_Entry->LFN_Name_7;
    cLongFileName[--i] = ptrLFN_Entry->LFN_Name_6;
    cLongFileName[--i] = ptrLFN_Entry->LFN_Name_5;
    cLongFileName[--i] = ptrLFN_Entry->LFN_Name_4;
    cLongFileName[--i] = ptrLFN_Entry->LFN_Name_3;
    cLongFileName[--i] = ptrLFN_Entry->LFN_Name_2;
    cLongFileName[--i] = ptrLFN_Entry->LFN_Name_1;
    cLongFileName[--i] = ptrLFN_Entry->LFN_Name_0;
    *ucLongFileNameLength = i;
}
#endif


// Read directory or file
//
static int utReadDirectory(UTLISTDIRECTORY *ptr_utListDirectory, UTFILEINFO *ptr_ut_fileInfo)
{
#ifdef UTFAT_LFN_READ
    CHAR cLongFileName[100];
    unsigned char ucLongFileNameLength = sizeof(cLongFileName);
#endif
    DIR_ENTRY_STRUCTURE_FAT32 *ptrDirectoryEntry;
    UTDISK *ptr_utDisk = &utDisks[ptr_utListDirectory->ptr_utDirObject->ucDrive]; // the disk that the directory is associated with
    int iFound = 0;

    while (iFound == 0) {
        if (fnLoadSector(ptr_utDisk, ptr_utListDirectory->private_disk_location.directory_location.ulSector) != UTFAT_SUCCESS) { // move to the disk sector containing the directory and read its content
            return UTFAT_DISK_READ_ERROR;
        }
        ptrDirectoryEntry = (DIR_ENTRY_STRUCTURE_FAT32 *)ptr_utDisk->ptrSectorData;
        ptrDirectoryEntry += ptr_utListDirectory->private_disk_location.ucDirectoryEntry; // move to the present entry
        switch (ptrDirectoryEntry->DIR_Name[0]) {
        case 0:                                                          // end of directory has been reached
            return UTFAT_END_OF_DIRECTORY;
        case DIR_NAME_FREE:                                              // 0xe5 means that the end of the directory has been reached
            break;
        default:
#ifdef UTFAT_LFN_READ                                                    // if long file name read support is enabled
            if ((ptrDirectoryEntry->DIR_Attr & DIR_ATTR_MASK) == (DIR_ATTR_LONG_NAME)) { // check whether the entry is a long file name
                fnExtractLongFileName(cLongFileName, (LFN_ENTRY_STRUCTURE_FAT32 *)ptrDirectoryEntry, &ucLongFileNameLength);
            }
            else if (!(ptrDirectoryEntry->DIR_Attr & DIR_ATTR_VOLUME_ID)) {  // if not a volume entry it is a short file name
                fnLoadShortFileInfo(ptr_ut_fileInfo, ptrDirectoryEntry); // extract the file information
                if (ucLongFileNameLength != sizeof(cLongFileName)) {
                    uStrcpy(ptr_ut_fileInfo->cFileName, &cLongFileName[ucLongFileNameLength]);
                }
                ucLongFileNameLength = sizeof(cLongFileName);            // reset long file name counter
                iFound = 1;                                              // mark that a file has been found and we can terminate after incrementing the directory entry
            }
#else
            if (!(ptrDirectoryEntry->DIR_Attr & DIR_ATTR_VOLUME_ID)) {   // if not a volume entry
                fnLoadShortFileInfo(ptr_ut_fileInfo, ptrDirectoryEntry); // extract the file information
                iFound = 1;                                              // mark that a file has been found and we can terminate after incrementing the directory entry
            }
#endif
            break;
        }
        if (fnNextDirectoryEntry(ptr_utDisk, &(ptr_utListDirectory->private_disk_location)) == UTFAT_DIRECTORY_AREA_EXHAUSTED) { // move to the next entry
            if (iFound == 1) {
                return UTFAT_FINAL_LISTING_ITEM_FOUND;
            }
            return UTFAT_DIRECTORY_AREA_EXHAUSTED;                       // quit since the complete directory space has been exhausted
        }
    }
    return UTFAT_SUCCESS;
}



// Perform a directory listing to a buffer (with options for DOS and FTP styles)
//
extern int utListDir(UTLISTDIRECTORY *ptr_utListDirectory, FILE_LISTING *ptrFileListingInfo)
{
    unsigned char ucLineLength;
    int iResult;
    CHAR *ptrBuf = ptrFileListingInfo->ptrBuffer;
    UTFILEINFO utFileInfo;                                               // temporary file info object
    DISK_LOCATION previous_disk_loc;
    unsigned short usYear;
    unsigned char  ucDay, ucMonth, ucHour, ucMinutes;
    ptrFileListingInfo->usStringLength = 0;
    if (ptrFileListingInfo->ucStyle == FTP_TYPE_LISTING) {
        ucLineLength = FTP_STYLE_LIST_ENTRY_LENGTH;                       // the maximum length of a single entry line - the user's buffer must be adequate to accept this
    }
    else {
        ucLineLength = DOS_STYLE_LIST_ENTRY_LENGTH;                       // the maximum length of a single entry line - the user's buffer must be adequate to accept this
    }
    ptrFileListingInfo->usItemsReturned = 0;
    while (ptrFileListingInfo->usItemsReturned < ptrFileListingInfo->usMaxItems) { // for maximum amount of items
        uMemcpy(&previous_disk_loc, &ptr_utListDirectory->private_disk_location, sizeof (ptr_utListDirectory->private_disk_location)); // backup in case we need to restore due to lack of buffer space
        iResult = utReadDirectory(ptr_utListDirectory, &utFileInfo);
        if ((iResult != UTFAT_SUCCESS) && (iResult != UTFAT_FINAL_LISTING_ITEM_FOUND)) { // read next directory or file in the present directory
            return UTFAT_NO_MORE_LISTING_ITEMS_FOUND;                    // no more items found
        }
        if ((ptrFileListingInfo->usStringLength + ucLineLength + uStrlen(utFileInfo.cFileName)) > ptrFileListingInfo->usBufferLength) {
            uMemcpy(&ptr_utListDirectory->private_disk_location, &previous_disk_loc, sizeof (ptr_utListDirectory->private_disk_location)); // restore location
            return UTFAT_NO_MORE_LISING_SPACE;                           // no more file listings fit in the present buffer
        }
        ucDay = (utFileInfo.usFileData & 0x1f);
        ucMonth = ((utFileInfo.usFileData >> 5) & 0xf);
        usYear = ((utFileInfo.usFileData >> 9) + 1980);
        ucMinutes = ((utFileInfo.usFileTime >> 5) & 0x3f);
        ucHour = (utFileInfo.usFileTime >> 11);
        if (ptrFileListingInfo->ucStyle == FTP_TYPE_LISTING) {
            int i = 10;
            unsigned short usRights;
            CHAR cAccess = 'd';
            if (utFileInfo.ucFileAttributes & DIR_ATTR_DIRECTORY) {
                usRights = 0x3ed;                                        // directory
            }
            else {                                                       // file
                if (utFileInfo.ucFileAttributes & DIR_ATTR_READ_ONLY) {
                    usRights = 0x124;
                }
                else {
                    usRights = 0x1a4;
                }
            }
            while (i--) {
                if (usRights & 0x200) {
                    *ptrBuf++ = cAccess;                                 // rights
                }
                else {
                    *ptrBuf++ = '-';                                     // no rights
                }
                switch (cAccess) {                                       // set next flag
                case 'r':
                    cAccess = 'w';
                    break;
                case 'w':
                    cAccess = 'x';
                    break;
                default:
                    cAccess = 'r';
                    break;
                }
                usRights <<= 1;
            }
            ptrBuf = uStrcpy(ptrBuf, " 1 502 502 ");
            ptrBuf = fnBufferDec(utFileInfo.ulFileSize, 0, ptrBuf);
            *ptrBuf++ = ' ';
            switch (ucMonth) {
            default:
                ptrBuf = uStrcpy(ptrBuf, "Jan");
                break;
            case 2:
                ptrBuf = uStrcpy(ptrBuf, "Feb");
                break;
            case 3:
                ptrBuf = uStrcpy(ptrBuf, "Mar");
                break;
            case 4:
                ptrBuf = uStrcpy(ptrBuf, "Apr");
                break;
            case 5:
                ptrBuf = uStrcpy(ptrBuf, "May");
                break;
            case 6:
                ptrBuf = uStrcpy(ptrBuf, "Jun");
                break;
            case 7:
                ptrBuf = uStrcpy(ptrBuf, "Jul");
                break;
            case 8:
                ptrBuf = uStrcpy(ptrBuf, "Aug");
                break;
            case 9:
                ptrBuf = uStrcpy(ptrBuf, "Sep");
                break;
            case 10:
                ptrBuf = uStrcpy(ptrBuf, "Oct");
                break;
            case 11:
                ptrBuf = uStrcpy(ptrBuf, "Nov");
                break;
            case 12:
                ptrBuf = uStrcpy(ptrBuf, "Dec");
                break;
            }
            ptrBuf = fnBufferDec(ucDay, (LEADING_ZERO | WITH_SPACE), ptrBuf); // day of month
            ptrBuf = fnBufferDec(usYear, WITH_SPACE, ptrBuf);
        }
        else {                                                           // DOS style listing
            if (utFileInfo.ucFileAttributes & DIR_ATTR_READ_ONLY) {
                *ptrBuf++ = 'R';
            }
            else {
                *ptrBuf++ = '-';
            }
            if (utFileInfo.ucFileAttributes & DIR_ATTR_HIDDEN) {
                *ptrBuf++ = 'H';
            }
            else {
                *ptrBuf++ = '-';
            }
            if (utFileInfo.ucFileAttributes & DIR_ATTR_SYSTEM) {
                *ptrBuf++ = 'S';
            }
            else {
                *ptrBuf++ = '-';
            }
            if (utFileInfo.ucFileAttributes & DIR_ATTR_ARCHIVE) {
                *ptrBuf++ = 'A';
            }
            else {
                *ptrBuf++ = '-';
            }
            ptrBuf = fnBufferDec(ucDay, (LEADING_ZERO | WITH_SPACE), ptrBuf);
            *ptrBuf++ = '.';
            ptrBuf = fnBufferDec(ucMonth, (LEADING_ZERO), ptrBuf);
            *ptrBuf++ = '.';
            ptrBuf = fnBufferDec(usYear, 0, ptrBuf);
            *ptrBuf++ = ' ';
            ptrBuf = fnBufferDec(ucHour, (LEADING_ZERO | WITH_SPACE), ptrBuf);
            *ptrBuf++ = ':';
            ptrBuf = fnBufferDec(ucMinutes, (LEADING_ZERO), ptrBuf);
            if (utFileInfo.ucFileAttributes & DIR_ATTR_DIRECTORY) {
                if (ptrFileListingInfo->ucStyle == DOS_TYPE_LISTING) {
                    ptrBuf = uStrcpy(ptrBuf, " <DIR>           ");
                }
                ptrFileListingInfo->usDirectoryCount++;                  // count the directories in this listing
            }
            else {
                int iLen;
                CHAR cLenBuf[18];                                        // maximum decimal file length plus null terminator plus " <DIR> "fill
                CHAR *ptrLen = fnBufferDec(utFileInfo.ulFileSize, WITH_SPACE, cLenBuf);
                iLen = (17 - (ptrLen - cLenBuf));
                while (iLen--) {
                    *ptrBuf++ = ' ';
                }
                ptrBuf = uStrcpy(ptrBuf, cLenBuf);
                ptrFileListingInfo->usFileCount++;                       // count the files in this listing
                ptrFileListingInfo->ulFileSizes += utFileInfo.ulFileSize;// sum of the total file sizes in this listing
            }
        }
        *ptrBuf++ = ' ';
        ptrBuf = uStrcpy(ptrBuf, utFileInfo.cFileName);
        *ptrBuf++ = '\r';
        *ptrBuf++ = '\n';
        ptrFileListingInfo->usStringLength = (ptrBuf - ptrFileListingInfo->ptrBuffer);
        ptrFileListingInfo->usItemsReturned++;
        if (iResult == UTFAT_FINAL_LISTING_ITEM_FOUND) {
            return UTFAT_NO_MORE_LISTING_ITEMS_FOUND;                    // the content is the final content
        }
    }
    return UTFAT_SUCCESS;
}


static int fnCreateNameParagraph(const CHAR **pptrDirectoryPath, CHAR cDirectoryName[12])
{
    int iLength = 0;
    int iEvenByte = 0;
    int iSeparatorFound = 0;
    CHAR cInputCharacter;
    unsigned char ucCharacterCharacteristics;

    uMemset(cDirectoryName, ' ', 11);                                    // start with blank name
    cDirectoryName[11] = NT_FLAG;
    while (1) {
        cInputCharacter = *(*pptrDirectoryPath)++;
        switch (cInputCharacter) {
        case '.':
            if ((iLength != 0) && (iLength <= 8)) {                      // separator found
                iSeparatorFound = 1;
                iLength = 8;
                continue;
            }
            break;
        case 0x0a:                                                       // for FTP compatibility
        case 0x0d:
            if (iLength != 0) {
                return FULLY_QUALIFIED_SHORT_NAME;                       // return true if the end of the path has been found
            }
            continue;
        case 0:
        case FORWARD_SLASH:
        case BACK_SLASH:
            if (iLength != 0) {
                return (cInputCharacter == 0);                           // return true if the end of the path has been found
            }                                                            // fall through to ignore
            continue;                                                    // ignore at start of string
        default:
            if ((unsigned char)cInputCharacter <= ' ') {                 // reject all invisible characters
                continue;
            }
            else {
                if ((unsigned char)cInputCharacter <= '~') {
                    ucCharacterCharacteristics = ucCharacterTable[cInputCharacter - '!'];
                    if (ucCharacterCharacteristics & (_CHAR_REJECT | _CHAR_REJECT_NON_JAP)) {
                        if ((iEvenByte == 0) || (!(ucCharacterCharacteristics & _CHAR_REJECT_NON_JAP))) { // don't apply these to second byte of Japanese characters
                            continue;                                    // reject
                        }
                    }
                    else if (iEvenByte == 0) {                           // don't apply to second byte of Japanese characters
                        if (ucCharacterCharacteristics & _CHAR_CAPITAL) {
                            if (iSeparatorFound != 1) {
                                cDirectoryName[11] &= ~0x08;
                            }
                            else {
                                cDirectoryName[11] &= ~0x10;
                            }
                        }
                        else if (ucCharacterCharacteristics & _CHAR_SMALL) {
                            cInputCharacter -= ('a' - 'A');              // convert to upper case
                            if (iSeparatorFound != 1) {
                                cDirectoryName[11] |= 0x08;
                            }
                            else {
                                cDirectoryName[11] |= 0x10;
                            }
                        }
                    }
                }
                else if ((((unsigned char)cInputCharacter >= 0x81) && ((unsigned char)cInputCharacter <= 0x9f)) || (((unsigned char)cInputCharacter >= 0xe0) && ((unsigned char)cInputCharacter <= 0xfc))) { // Japanese
                    iEvenByte = 1;
                    if ((iLength == 0) && (cInputCharacter == DIR_NAME_FREE)) {
                        cDirectoryName[iLength++] = 0x05;                // substitute
                        continue;                                        // continue with even byte
                    }
                }
                else {
                    continue;                                            // reject
                }
            }
            break;                                                       // accept
        }
        if (iLength >= 8) {
            if ((iSeparatorFound == 0) || (iLength >= 11)) {             // name part or extension is too long
#ifdef UTFAT_LFN_READ
                while (1) {
                    switch (*(*pptrDirectoryPath)++) {                   // search the end of the long file name paragraph
                    case 0x0d:                                           // end of line considered fully qualified terminator (for FTP compatibility)
                    case 0x0a:
                    case 0:
                        (*pptrDirectoryPath)--;                          // set back to null-terinator of long file name
                        return FULLY_QUALIFIED_LONG_NAME;
                    case FORWARD_SLASH:
                    case BACK_SLASH:
                        (*pptrDirectoryPath)--;                          // set back to terminator of long file name
                        return LONG_NAME_PARAGRAPH;
                    }
                }
#else
                break;
#endif
            }
        }
        iEvenByte = 0;
        cDirectoryName[iLength++] = cInputCharacter;
    }
    return INVALID_PARAGRAPH;                                            // invalid
}


static UTDIRECTORY utDirectoryObjects[UT_DIRECTORIES_AVAILABLE] = {{0}};

extern UTDIRECTORY*utAllocateDirectory(unsigned char ucDisk, unsigned short usPathLength)
{
    int i = 0;
    while (i < UT_DIRECTORIES_AVAILABLE) {
        if (utDirectoryObjects[i].ucDirectoryFlags == 0) {               // directory not allocated
            utDirectoryObjects[i].ucDirectoryFlags = UTDIR_ALLOCATED;
            utDirectoryObjects[i].ucDrive = ucDisk;
            if (usPathLength != 0) {
                utDirectoryObjects[i].ptrDirectoryPath = uMalloc((MAX_MALLOC)(usPathLength + 1)); // reserve space for holding the directory path string (plus space for null terminator)
            }
            else {
                utDirectoryObjects[i].ptrDirectoryPath = 0;
            }
            utDirectoryObjects[i].usDirectoryPathLength = usPathLength;  // enter the maximum length that can be stored
            return &utDirectoryObjects[i];
        }
        i++;
    }
    return 0;                                                            // no directory object available for allocation
}


// Open or modify a directory object
//
extern int utOpenDirectory(const CHAR *ptrDirPath, UTDIRECTORY *ptrDirObject)
{
    DIR_ENTRY_STRUCTURE_FAT32 *ptrFoundEntry = 0;
    UTDISK *ptr_utDisk = &utDisks[ptrDirObject->ucDrive];
    DISK_LOCATION *ptrDiskLocation;
    int iRootDirectory = 0;
    const CHAR *ptrLocalDirPath = ptrDirPath;
#ifdef UTFAT_LFN_READ
    const CHAR *ptrLongFileNamePathStart, *ptrLongFileNamePathEnd;
#endif
    unsigned long ulCluster;
    DISK_LOCATION DirStart;
    int iMatchedParagraph;
    int iFullyQualifiedPath;
    CHAR cFileName[8 + 3 + 1];                                           // space for 8:3 format plus an NT specific byte
    unsigned char ucDirFlags = ptrDirObject->ucDirectoryFlags;
    ptrDirObject->ucDirectoryFlags &= ~(UTDIR_TEST_FULL_PATH | UTDIR_TEST_FULL_PATH_TEMP | UTDIR_TEST_REL_PATH | UTDIR_DIR_AS_FILE | UTDIR_SET_START); // these flags are automatically cleared

    if (!(ptr_utDisk->ucDiskFlags & DISK_MOUNTED)) {                     // if the disk is not ready for use quit
        return UTFAT_DISK_NOT_READY;
    }
    if ((ptrLocalDirPath == 0) || (*ptrLocalDirPath == 0) || ((*ptrLocalDirPath == LINE_FEED) || (*ptrLocalDirPath == CARRIAGE_RETURN)) || (((*ptrLocalDirPath == BACK_SLASH) || (*ptrLocalDirPath == FORWARD_SLASH)) && (*(ptrLocalDirPath + 1) == 0))) { // if root directory
        iRootDirectory = 1;                                              // ready to work with root directory or referenced directory
    }

    if (!(ucDirFlags & UTDIR_VALID)) {
        ulCluster = ptr_utDisk->ulDirectoryBase;                         // the first cluster in the root directory
        ptrDirObject->private_disk_location.directory_location.ulCluster = ulCluster;
        ptrDirObject->private_disk_location.directory_location.ulSector = (ulCluster * ptr_utDisk->utFAT.ucSectorsPerCluster);// section referenced to logical base address
        ptrDirObject->private_disk_location.directory_location.ulSector += ptr_utDisk->ulVirtualBaseAddress;// the sector in which the directory entries begin
        ptrDirObject->private_disk_location.ucDirectoryEntry = 0;        // reset the entry index
        ptrDirObject->ucDirectoryFlags |= UTDIR_VALID;                   // the entry is now valid
        if (iRootDirectory != 0) {                                       // first open of root directory so enter the root path
            if (ptrDirObject->ptrDirectoryPath != 0) {                   // if a root path is being used
                uStrcpy(ptrDirObject->ptrDirectoryPath, "D:\\");
                ptrDirObject->usRelativePathLocation = 3;                // the relative path is equal to the root path
            }
        }
    }

    if (ucDirFlags & (UTDIR_TEST_FULL_PATH | UTDIR_TEST_FULL_PATH_TEMP | UTDIR_TEST_REL_PATH | UTDIR_REFERENCED)) {
        if (ucDirFlags & UTDIR_TEST_FULL_PATH) {                         // set temporary root directory
            ulCluster = ptr_utDisk->ulDirectoryBase;                     // the first cluster in the root directory
            ptrDirObject->public_disk_location.directory_location.ulCluster = ulCluster;
            ptrDirObject->public_disk_location.directory_location.ulSector = (ulCluster * ptr_utDisk->utFAT.ucSectorsPerCluster);
            ptrDirObject->public_disk_location.directory_location.ulSector += ptr_utDisk->ulVirtualBaseAddress;
            ptrDirObject->public_disk_location.ucDirectoryEntry = 0;
        }
        else if (!(ucDirFlags & UTDIR_TEST_FULL_PATH_TEMP)) {
            uMemcpy(&ptrDirObject->public_disk_location, &ptrDirObject->private_disk_location, sizeof(ptrDirObject->private_disk_location)); // synchronise
        }                                                                // UTDIR_TEST_FULL_PATH_TEMP continues using present public values to continue a search
        ptrDiskLocation = &ptrDirObject->public_disk_location;        
    }
    else {
        ptrDiskLocation = &ptrDirObject->private_disk_location;
    }
    if (iRootDirectory != 0) {                                           // if root directory
        return UTFAT_SUCCESS;                                            // ready to work with root directory or referenced directory
    }
    uMemcpy(&DirStart, ptrDiskLocation, sizeof(DirStart));               // {6} backup the original starting directory location

    while (1) {
#ifdef UTFAT_LFN_READ
        ptrLongFileNamePathStart = ptrLocalDirPath;
        if ((*ptrLongFileNamePathStart == BACK_SLASH) || (*ptrLongFileNamePathStart == FORWARD_SLASH)) {
            ptrLongFileNamePathStart++;
        }
#endif
        if ((iFullyQualifiedPath = fnCreateNameParagraph(&ptrLocalDirPath, cFileName)) < 0) { // fill the file name with a single paragraph from the file path
            return UTFAT_PATH_NOT_FOUND;                                                   //
        }
#ifdef UTFAT_LFN_READ
        ptrLongFileNamePathEnd = ptrLocalDirPath;
#endif
        iMatchedParagraph = 0;
        do {
            if (fnLoadSector(ptr_utDisk, ptrDiskLocation->directory_location.ulSector) != UTFAT_SUCCESS) {
                return UTFAT_DISK_READ_ERROR;
            }
            ptrFoundEntry = (DIR_ENTRY_STRUCTURE_FAT32 *)ptr_utDisk->ptrSectorData; // the directory entry in the sector buffer
            ptrFoundEntry += ptrDiskLocation->ucDirectoryEntry;          // move to the present entry
#ifdef UTFAT_LFN_READ
            if (iFullyQualifiedPath > FULLY_QUALIFIED_SHORT_NAME) {      // the name is not a short name so check for long name match
                int iMatchStatus = fnMatchLongName(&ptrLongFileNamePathEnd, ptrLongFileNamePathStart, (LFN_ENTRY_STRUCTURE_FAT32 *)ptrFoundEntry);
                switch (iMatchStatus) {
                case MATCH_SUCCESSFUL:
                    if (iFullyQualifiedPath == FULLY_QUALIFIED_LONG_NAME) {
                        iFullyQualifiedPath = 1;                         // path name fully matched
                    }
                    else {
                        iFullyQualifiedPath = 0;                         // successful paragraph match, but not completed
                    }
                    ptrLongFileNamePathStart = 0;                        // for an unconditional match on next directory entry
                    break;
                case MATCH_FALSE:
                    ptrLongFileNamePathEnd = ptrLocalDirPath;
                case MATCH_CONTINUE:                                     // match is still correct but not completed
                    break;                                               // not the searched name
                case END_DIRECTORY_ENTRIES:                              // end if end found
                    if (iFullyQualifiedPath == FULLY_QUALIFIED_LONG_NAME) {
                        return UTFAT_FILE_NOT_FOUND;
                    }
                    else {
                        return UTFAT_PATH_NOT_FOUND;
                    }
                }
            }
            else {
#endif
                switch (ptrFoundEntry->DIR_Name[0]) {
                case 0:                                                  // end if end found
                    if (iFullyQualifiedPath != 0) {
                        if (ucDirFlags & UTDIR_SET_START) {              // {6}
                            uMemcpy(ptrDiskLocation, &DirStart, sizeof(DirStart)); // set the location to the start of the lowest directory
                        }
                        return UTFAT_FILE_NOT_FOUND;
                    }
                    else {
                        return UTFAT_PATH_NOT_FOUND;
                    }
                case DIR_NAME_FREE:
                    break;
                default:
                    if (
#ifdef UTFAT_LFN_READ
                        (ptrLongFileNamePathStart == 0) ||               // this short file name entry unconditionally belongs to a previous long file name block
#endif
                        ((!(ptrFoundEntry->DIR_Attr & DIR_ATTR_VOLUME_ID)) && (!uMemcmp(ptrFoundEntry->DIR_Name, cFileName, sizeof(ptrFoundEntry->DIR_Name))))) { // if entry has been found
                        ulCluster = ((ptrFoundEntry->DIR_FstClusHI[1] << 24) + (ptrFoundEntry->DIR_FstClusHI[0] << 16) + (ptrFoundEntry->DIR_FstClusLO[1] << 8) + ptrFoundEntry->DIR_FstClusLO[0]);
                        ptrDirObject->ptrEntryStructure = ptrFoundEntry; // set pointer to the entry so that the caller can extract additional information if required
                        if (!(ptrFoundEntry->DIR_Attr & DIR_ATTR_DIRECTORY)) {
                            ptrDirObject->public_file_location.ulCluster = ulCluster; // file's start cluster
                            ptrDirObject->public_file_location.ulSector = ((ulCluster * ptr_utDisk->utFAT.ucSectorsPerCluster) + ptr_utDisk->ulVirtualBaseAddress);// section referenced to logical base address
                            return UTFAT_PATH_IS_FILE;                   // not a directory so can not be traced further
                        }
                        if ((iFullyQualifiedPath != 0) && (ucDirFlags & UTDIR_DIR_AS_FILE)) { // if a directory is to be treated as a file, its location is preserved rather than moving to its content
                            return UTFAT_SUCCESS;                        // file matched
                        }
                        ptrDiskLocation->directory_location.ulCluster = ulCluster; // move to the new directory
                        ptrDiskLocation->directory_location.ulSector = ((ulCluster * ptr_utDisk->utFAT.ucSectorsPerCluster) + ptr_utDisk->ulVirtualBaseAddress);// section referenced to logical base address
                        ptrDiskLocation->ucDirectoryEntry = 2;           // skip "." and ".." entries
                        if (iFullyQualifiedPath != 0) {
                            return UTFAT_SUCCESS;                        // file matched
                        }
                        uMemcpy(&DirStart, ptrDiskLocation, sizeof(DirStart)); // {6} set the location to the start of this directory
                        iMatchedParagraph = 1;                           // this paragraph matched, but not last one
                    }
                    break;
                }
#ifdef UTFAT_LFN_READ
            }
#endif
            if (iMatchedParagraph != 0) {
                break;
            }
            if (fnNextDirectoryEntry(ptr_utDisk, ptrDiskLocation) != UTFAT_SUCCESS) { // move to next directory entry
                if (iMatchedParagraph != 0) {
                    return UTFAT_SUCCESS;
                }
                if (iFullyQualifiedPath != 0) {
                    return UTFAT_FILE_NOT_FOUND;
                }
                else {
                    return UTFAT_PATH_NOT_FOUND;
                }
            }
        } while (1);
    }
}



// Read linear file data content directly to an external buffer
//
extern int utReadFile(UTFILE *ptr_utFile, void *ptrBuffer, unsigned short usReadLength)
{
    UTDISK *ptr_utDisk = &utDisks[ptr_utFile->ucDrive];
    unsigned long ulContentRemaining;
    unsigned short usAccessLength;
    unsigned short usAccessOffset;

    ptr_utFile->usLastReadWriteLength = 0;
    if (!(ptr_utFile->ucFileMode & UTFAT_OPEN_FOR_READ)) {
        return UTFAT_FILE_NOT_READABLE;                                  // don't allow read of files not opened for read
    }
    ulContentRemaining = (ptr_utFile->ulFileSize - ptr_utFile->ulFilePosition);
    if (usReadLength > ulContentRemaining) {
        usReadLength = (unsigned short)ulContentRemaining;               // limit  to remaining size
    }

    while (usReadLength != 0) {                                          // while requested length of data has not yet been returned
        usAccessOffset = (unsigned short)(ptr_utFile->ulFilePosition & 0x1ff);
        usAccessLength = (512 - usAccessOffset);
        if (usAccessLength > usReadLength) {
            usAccessLength = usReadLength;
        }
        if (fnLoadData(ptr_utDisk, ptr_utFile->public_file_location.ulSector, (unsigned char *)ptrBuffer, usAccessOffset, usAccessLength) != UTFAT_SUCCESS) { // read directly to buffer
            return UTFAT_DISK_READ_ERROR;
        }
        usReadLength -= usAccessLength;
        ptr_utFile->ulFilePosition += usAccessLength;
        ptr_utFile->usLastReadWriteLength += usAccessLength;
        ptrBuffer = (void *)((CAST_POINTER_ARITHMETIC)ptrBuffer + usAccessLength);
        if (((ptr_utFile->ulFilePosition % ptr_utDisk->utFAT.usBytesPerSector) == 0) && (ptr_utFile->ulFileSize > ptr_utFile->ulFilePosition)) { // on a sector boundary when file end has not already been reached
            int iResult = fnNextSector(ptr_utDisk, &ptr_utFile->public_file_location);
            if (iResult != UTFAT_SUCCESS) {
                return iResult;
            }
        }
    }
    return UTFAT_SUCCESS;
}


// Open a list directory referenced to its main directory object, which must already have been configured
//
extern int utLocateDirectory(const CHAR *ptrDirPath, UTLISTDIRECTORY *ptrListDirectory)
{
    int iReturn = UTFAT_SUCCESS;
    if (ptrListDirectory->ptr_utDirObject == 0) {
        return UTFAT_DIRECTORY_OBJECT_MISSING;
    }
    ptrListDirectory->ptr_utDirObject->ucDirectoryFlags |= UTDIR_REFERENCED; // the next open is a referenced open
    if ((iReturn = utOpenDirectory(ptrDirPath, ptrListDirectory->ptr_utDirObject)) == UTFAT_SUCCESS) {
        uMemcpy(&ptrListDirectory->private_disk_location, &ptrListDirectory->ptr_utDirObject->public_disk_location, sizeof(ptrListDirectory->private_disk_location)); // copy the referenced directory details to the list directory object
    }
    return iReturn;
}

// This routine sets a fixed time and data - it can be replaced by a routine to set the infomrtaion from a RTC or similar
//
#define CREATION_HOURS   15
#define CREATION_MINUTES 04
#define CREATION_SECONDS 25

#define CREATION_DAY_OF_MONTH  20
#define CREATION_MONTH_OF_YEAR 1                                         // january
#define CREATION_YEAR    (2010 - 1980)

static void fnSetTimeDate(DIR_ENTRY_STRUCTURE_FAT32 *ptrEntry, int iCreation) // {3}
{
    unsigned short usCreationTime = (CREATION_SECONDS | (CREATION_MINUTES << 5) | (CREATION_HOURS << 11));
    unsigned short usCreationDate = (CREATION_DAY_OF_MONTH | (CREATION_MONTH_OF_YEAR << 5) | (CREATION_YEAR << 9));
    ptrEntry->DIR_WrtTime[0] = (unsigned char)(usCreationTime);
    ptrEntry->DIR_WrtTime[1] = (unsigned char)(usCreationTime >> 8);
    ptrEntry->DIR_LstAccDate[0] = ptrEntry->DIR_WrtDate[0] = (unsigned char)(usCreationDate);
    ptrEntry->DIR_LstAccDate[1] = ptrEntry->DIR_WrtDate[1] = (unsigned char)(usCreationDate >> 8);
    if (iCreation != 0) {
        ptrEntry->DIR_CrtTime[0] = ptrEntry->DIR_WrtTime[0];
        ptrEntry->DIR_CrtTime[1] = ptrEntry->DIR_WrtTime[1];
        ptrEntry->DIR_CrtDate[0] = ptrEntry->DIR_LstAccDate[0];
        ptrEntry->DIR_CrtDate[1] = ptrEntry->DIR_LstAccDate[1];
    }
}

static void fnAddEntry(DIR_ENTRY_STRUCTURE_FAT32 *ptrEntry, unsigned long ulPresentCluster, unsigned char ucAttributes)
{
    ptrEntry->DIR_FstClusLO[0] = (unsigned char)ulPresentCluster;
    ptrEntry->DIR_FstClusLO[1] = (unsigned char)(ulPresentCluster >> 8);
    ptrEntry->DIR_FstClusHI[0] = (unsigned char)(ulPresentCluster >> 16);
    ptrEntry->DIR_FstClusHI[1] = (unsigned char)(ulPresentCluster >> 24);
    ptrEntry->DIR_Attr = ucAttributes;
    uMemset(ptrEntry->DIR_FileSize, 0, sizeof(ptrEntry->DIR_FileSize));  // {2} set the file size to zero
    fnSetTimeDate(ptrEntry, 1);                                          // {3} add creation time
}

static void fnSetFileInformation(DIR_ENTRY_STRUCTURE_FAT32 *ptrFileEntry, unsigned long ulFileSize)
{
    ptrFileEntry->DIR_FileSize[0] = (unsigned char)(ulFileSize);         // update the file size
    ptrFileEntry->DIR_FileSize[1] = (unsigned char)(ulFileSize >> 8);
    ptrFileEntry->DIR_FileSize[2] = (unsigned char)(ulFileSize >> 16);
    ptrFileEntry->DIR_FileSize[3] = (unsigned char)(ulFileSize >> 24);
    fnSetTimeDate(ptrFileEntry, 0);                                      // {3} set the modification time
}

// Allocate a single new cluster
//
static unsigned long fnAllocateCluster(UTDISK *ptr_utDisk, unsigned long ulPresentCluster, unsigned char ucClusterType)
{
    unsigned long ulFAT = ptr_utDisk->utFAT.ulFAT_start;
    unsigned long ulFatOriginal = ulFAT;
    unsigned long ulSectorContent[512/sizeof(signed long)];              // sector to read FAT32 sectors to
    unsigned long ulAbsoluteCluster;
    unsigned char ucClusterEntry;
    unsigned char ucOriginalCluster;
    if (ucClusterType & NEW_RELATIVE_CLUSTER) {
        ulAbsoluteCluster = ulPresentCluster;
        ulFAT += (ulPresentCluster/(512/sizeof(signed long)));           // {4}
    }
    else {
        if ((ptr_utDisk->ucDiskFlags & FSINFO_VALID) && (ptr_utDisk->utFileInfo.ulNextFreeCluster <= ptr_utDisk->utFileInfo.ulFreeClusterCount)) { // the info block is valid so reference to next free cluster
            ulAbsoluteCluster = ptr_utDisk->utFileInfo.ulNextFreeCluster;
        }
        else {
            ulAbsoluteCluster = ptr_utDisk->ulDirectoryBase;
        }
        ulPresentCluster = ulAbsoluteCluster;
    }
    ucClusterEntry = (unsigned char)ulPresentCluster;
    ucClusterEntry &= ((512/sizeof(signed long)) - 1);
    ucOriginalCluster = ucClusterEntry;
    while (1) {
        if ((utReadDiskData(ptr_utDisk, ulFAT, ulSectorContent, 0, 512)) != UTFAT_SUCCESS) { // read a FAT sector containing the cluster information
            return UTFAT_DISK_READ_ERROR;
        }
        while (ucClusterEntry < (512/sizeof(unsigned long))) {
            if (ulSectorContent[ucClusterEntry] == 0) {                  // next free cluster location found
                unsigned char ucFatCopies = 0;
                if (ucClusterType & (INITIALISE_DIR_CLUSTER | INITIALISE_DIR_EXTENSION)) { // the new cluster must be configured as empty directory
                    unsigned char ucSectors = ptr_utDisk->utFAT.ucSectorsPerCluster;
                    unsigned long ulSectorToDelete = ((ulAbsoluteCluster * ptr_utDisk->utFAT.ucSectorsPerCluster) + ptr_utDisk->ulVirtualBaseAddress);
                    while (ucSectors != 0) {
                        if ((ucSectors-- == ptr_utDisk->utFAT.ucSectorsPerCluster) && (ucClusterType & INITIALISE_DIR_CLUSTER)) { // first sector - add "." and ".." entries
                            unsigned char ucDirectoryDefault[512];
                            DIR_ENTRY_STRUCTURE_FAT32 *ptrEntry = (DIR_ENTRY_STRUCTURE_FAT32 *)ucDirectoryDefault;
                            uMemset(ucDirectoryDefault, 0x00, sizeof(ucDirectoryDefault));
                            uMemset(ptrEntry->DIR_Name, ' ', sizeof(ptrEntry->DIR_Name));
                            ptrEntry->DIR_Name[0] = '.';
                            fnAddEntry(ptrEntry, ulPresentCluster, DIR_ATTR_DIRECTORY);
                            ptrEntry++;
                            uMemset(ptrEntry->DIR_Name, ' ', sizeof(ptrEntry->DIR_Name));
                            ptrEntry->DIR_Name[0] = '.';
                            ptrEntry->DIR_Name[1] = '.';
                            fnAddEntry(ptrEntry, 2, DIR_ATTR_DIRECTORY);
                            while (utCommitSectorData(ptr_utDisk, ucDirectoryDefault, ulSectorToDelete) == CARD_BUSY_WAIT) {}
                        }
                        else {
                            while (utDeleteSector(ptr_utDisk, ulSectorToDelete) == CARD_BUSY_WAIT) {} // delete sector
                        }
                        ulSectorToDelete++;
                    }
                }
                ulSectorContent[ucClusterEntry] = LITTLE_LONG_WORD(CLUSTER_MASK); // mark last cluster in extension
                if (ucClusterType & UPDATE_FAT_END) {
                    ucClusterType = 0;
                    ulSectorContent[ucOriginalCluster] = LITTLE_LONG_WORD(ulAbsoluteCluster); // modify the original end marker to point to the additional cluster
                }
                while (ucFatCopies < ptr_utDisk->utFAT.ucNumberOfFATs) {
                    while (utCommitSectorData(ptr_utDisk, ulSectorContent, (ulFAT + (ucFatCopies * ptr_utDisk->utFAT.ulFatSize))) == CARD_BUSY_WAIT) {} // write the new FAT entry
                    ucFatCopies++;
                }
                if (ucClusterType & UPDATE_FAT_END_IN_DIFF_SECT) {       // the new cluster end has been marked but the original end must be modified to point to it (it is in a different sector so needs to be modified seperately)
                    ucFatCopies = 0;
                    if ((utReadDiskData(ptr_utDisk, ulFatOriginal, ulSectorContent, 0, 512)) != UTFAT_SUCCESS) { // read a FAT sector containing the cluster information
                        return UTFAT_DISK_READ_ERROR;
                    }
                    ulSectorContent[ulPresentCluster & ((512/sizeof(signed long)) - 1)] = LITTLE_LONG_WORD(ulAbsoluteCluster);
                    while (ucFatCopies < ptr_utDisk->utFAT.ucNumberOfFATs) {
                        while (utCommitSectorData(ptr_utDisk, ulSectorContent, (ulFatOriginal + (ucFatCopies * ptr_utDisk->utFAT.ulFatSize))) == CARD_BUSY_WAIT) {} // write the new FAT entry
                        ucFatCopies++;
                    }
                }
                if (ptr_utDisk->ucDiskFlags & FSINFO_VALID) {
                    if (ulAbsoluteCluster >= ptr_utDisk->utFileInfo.ulNextFreeCluster) {                        
                        ptr_utDisk->utFileInfo.ulNextFreeCluster = (ulAbsoluteCluster + 1);
                    }
                    ptr_utDisk->ucDiskFlags |= WRITEBACK_INFO_FLAG;      // mark that the info block information has changed
                    ptr_utDisk->utFileInfo.ulFreeClusterCount--;
                }
                return (ulAbsoluteCluster);
            }
            ucClusterEntry++;
            ulAbsoluteCluster++;
        }
        ucClusterType &= ~UPDATE_FAT_END;                                // the new FAT entry is in a different FAT sector so needs to be modified later
        ucClusterEntry = 0;
        ulFAT++;
    }
}


static void fnAddInfoSect(INFO_SECTOR_FAT32 *ptrInfoSector, unsigned long ulFreeCount, unsigned long ulNextFree)
{
    ptrInfoSector->FSI_Free_Count[0] = (unsigned char)(ulFreeCount);
    ptrInfoSector->FSI_Free_Count[1] = (unsigned char)(ulFreeCount >> 8);
    ptrInfoSector->FSI_Free_Count[2] = (unsigned char)(ulFreeCount >> 16);
    ptrInfoSector->FSI_Free_Count[3] = (unsigned char)(ulFreeCount >> 24);
    ptrInfoSector->FSI_Nxt_Free[0]   = (unsigned char)(ulNextFree);
    ptrInfoSector->FSI_Nxt_Free[1]   = (unsigned char)(ulNextFree >> 8);
    ptrInfoSector->FSI_Nxt_Free[2]   = (unsigned char)(ulNextFree >> 16);
    ptrInfoSector->FSI_Nxt_Free[3]   = (unsigned char)(ulNextFree >> 24);
}

static int fnCommitInfoChanges(UTDISK *ptr_utDisk)
{
    if (ptr_utDisk->ucDiskFlags & WRITEBACK_INFO_FLAG) {                 // info sector content has changed
        INFO_SECTOR_FAT32 info_Sector;
        unsigned long ulInfoSector_Location = ptr_utDisk->utFileInfo.ulInfoSector;
        if ((utReadDiskData(ptr_utDisk, ulInfoSector_Location, &info_Sector, 0, sizeof(INFO_SECTOR_FAT32))) != UTFAT_SUCCESS) { // read the information sector
            return UTFAT_DISK_READ_ERROR;
        }
        fnAddInfoSect(&info_Sector, ptr_utDisk->utFileInfo.ulFreeClusterCount, ptr_utDisk->utFileInfo.ulNextFreeCluster);
        while (utCommitSectorData(ptr_utDisk, &info_Sector, ulInfoSector_Location) == CARD_BUSY_WAIT) {}
        ptr_utDisk->ucDiskFlags &= ~WRITEBACK_INFO_FLAG;
    }
    return UTFAT_SUCCESS;
}

// Open a list directory referenced to its main directory object, which must already have been configured
//
extern int utOpenFile(const CHAR *ptrFilePath, UTFILE *ptr_utFile, unsigned char ucAccessMode)
{
    int iReturn = UTFAT_SUCCESS;
    if (ptr_utFile->ptr_utDirObject == 0) {
        return UTFAT_SEARCH_INVALID;
    }
    if (ucAccessMode & UTFAT_MANAGED_MODE) {
        int iFileHandle = 0;
        UTMANAGED_FILE *ptrManagedFile = utManagedFiles;
        while (iFileHandle < UTMANAGED_FILE_COUNT) {
            if (ptrManagedFile->managed_owner == 0) {                
                if ((iReturn = utOpenFile(ptrFilePath, ptr_utFile, (unsigned char)(ucAccessMode & ~UTFAT_MANAGED_MODE)) != UTFAT_PATH_IS_FILE)) {
                    return UTFAT_FILE_NOT_FOUND;                         // file doesn't exist
                }
                if (fnFileLocked(ptr_utFile) != 0) {
                    return UTFAT_FILE_LOCKED;                            // the file is locked by another user
                }
                ptrManagedFile->utManagedFile = ptr_utFile;
                ptrManagedFile->managed_owner = ptr_utFile->ownerTask;   // validate the entry with the owner task
                ptr_utFile->ucFileMode = ucAccessMode;
                ptrManagedFile->managed_mode = ucAccessMode;
                ptr_utFile->iFileHandle = iFileHandle;
                return UTFAT_PATH_IS_FILE;                               // open in managed mode was successful
            }
            iFileHandle++;
            ptrManagedFile++;
        }
        return MANAGED_FILE_NO_FILE_HANDLE;                              // all managed file spaces are presently occupied
    }
    ptr_utFile->ptr_utDirObject->ucDirectoryFlags |= (UTDIR_REFERENCED | UTDIR_SET_START); // {6} the next open is a referenced open; set to start o lowest directory when file not found
    if ((iReturn = utOpenDirectory(ptrFilePath, ptr_utFile->ptr_utDirObject)) == UTFAT_PATH_IS_FILE) {
        uMemcpy(&ptr_utFile->private_disk_location, &ptr_utFile->ptr_utDirObject->public_disk_location, sizeof(ptr_utFile->private_disk_location)); // copy the referenced directory details
        uMemcpy(&ptr_utFile->private_file_location, &ptr_utFile->ptr_utDirObject->public_file_location, sizeof(ptr_utFile->private_file_location)); // copy the referenced file start details
        uMemcpy(&ptr_utFile->public_file_location, &ptr_utFile->ptr_utDirObject->public_file_location, sizeof(ptr_utFile->public_file_location)); // copy the referenced file start details
        ptr_utFile->ucFileMode = ucAccessMode;
        ptr_utFile->ulFileSize = ((ptr_utFile->ptr_utDirObject->ptrEntryStructure->DIR_FileSize[3] << 24) + (ptr_utFile->ptr_utDirObject->ptrEntryStructure->DIR_FileSize[2] << 16) + (ptr_utFile->ptr_utDirObject->ptrEntryStructure->DIR_FileSize[1] << 8) + ptr_utFile->ptr_utDirObject->ptrEntryStructure->DIR_FileSize[0]);
        ptr_utFile->ulFilePosition = 0;                                  // set position to start of file on open
        ptr_utFile->ucDrive = ptr_utFile->ptr_utDirObject->ucDrive;
#ifdef UTFAT_WRITE
        if (ucAccessMode & UTFAT_TRUNCATE) {                             // the file is to be overwritten so delete its content
            fnDeleteFileContent(ptr_utFile, &utDisks[ptr_utFile->ucDrive], 0);
            ptr_utFile->ulFileSize = 0;
        }
#endif
        return (UTFAT_PATH_IS_FILE);
    }
#ifdef UTFAT_WRITE
    else if (ucAccessMode & UTFAT_CREATE) {                              // file doesn't exist so we should create it
        const CHAR *ptrFilename = ptrFilePath;
        DISK_LOCATION *ptrDiskLocation = &ptr_utFile->ptr_utDirObject->public_disk_location; // this is the start of the directory in which th enew file is to be placed - this allows reuse of deleted file locations in the directory
        DIR_ENTRY_STRUCTURE_FAT32 *ptrFoundEntry;
        UTDISK *ptr_utDisk = &utDisks[ptr_utFile->ptr_utDirObject->ucDrive];
        CHAR cFileName[8 + 3 + 1];                                       // space for 8:3 format plus an NT specific byte
        if (iReturn != UTFAT_FILE_NOT_FOUND) {                           // invalid path has been entered so return
            return iReturn;
        }
        do {
            iReturn = fnCreateNameParagraph(&ptrFilename, cFileName);    // run down the path until the file name is loaded
        } while ((iReturn != FULLY_QUALIFIED_SHORT_NAME) && (iReturn != FULLY_QUALIFIED_LONG_NAME));
        while (1) {
            if (fnLoadSector(ptr_utDisk, ptrDiskLocation->directory_location.ulSector) != UTFAT_SUCCESS) {
                return UTFAT_DISK_READ_ERROR;
            }
            ptrFoundEntry = (DIR_ENTRY_STRUCTURE_FAT32 *)ptr_utDisk->ptrSectorData; // the directory entry in the sector buffer
            ptrFoundEntry += ptrDiskLocation->ucDirectoryEntry;          // move to the present entry
            if ((DIR_NAME_FREE == ptrFoundEntry->DIR_Name[0]) || (ptrFoundEntry->DIR_Name[0] == 0)) { // deleted entry of end of root entries found
                unsigned long ulFreeCluster = fnAllocateCluster(ptr_utDisk, 0, 0);
                uMemcpy(ptrFoundEntry->DIR_Name, cFileName, 11);         // add the short name
                ptrFoundEntry->DIR_NTRes = cFileName[11];
                fnAddEntry(ptrFoundEntry, ulFreeCluster, DIR_ATTR_ARCHIVE);
                while (utCommitSector(ptr_utDisk, ptrDiskLocation->directory_location.ulSector) == CARD_BUSY_WAIT) {} // force writeback to finalise the operation
                ptr_utDisk->ucDiskFlags &= ~WRITEBACK_BUFFER_FLAG;
                uMemcpy(&ptr_utFile->private_disk_location, &ptr_utFile->ptr_utDirObject->public_disk_location, sizeof(ptr_utFile->private_disk_location)); // copy the referenced directory details
                ptr_utFile->private_file_location.ulCluster = ulFreeCluster;
                ptr_utFile->private_file_location.ulSector = (ulFreeCluster * ptr_utDisk->utFAT.ucSectorsPerCluster);// section referenced to logical base address
                ptr_utFile->private_file_location.ulSector += ptr_utDisk->ulVirtualBaseAddress;// the sector in which the file content begins
                uMemcpy(&ptr_utFile->public_file_location, &ptr_utFile->private_file_location, sizeof(ptr_utFile->private_file_location)); // copy the referenced file start details
                ptr_utFile->ucFileMode = ucAccessMode;
                ptr_utFile->ulFileSize = 0;
                ptr_utFile->ulFilePosition = 0;                          // set position to start of file on open
                ptr_utFile->ucDrive = ptr_utFile->ptr_utDirObject->ucDrive;
                fnCommitInfoChanges(ptr_utDisk);
                return UTFAT_PATH_IS_FILE;
            }
            else {
                if (fnNextDirectoryEntry(ptr_utDisk, ptrDiskLocation) != UTFAT_SUCCESS) { // move to next directory entry
                    int iResult;
                    iResult = fnDirectorySectorCreate(ptr_utDisk, &ptrDiskLocation->directory_location); // move to next sector/cluster and create additional cluster if necessary
                    if (iResult != UTFAT_SUCCESS) {
                        return iResult;
                    }
                }
            }
        }
    }
#endif
    else if (iReturn == UTFAT_SUCCESS) {                                 // a directory was matched 
        uMemcpy(&ptr_utFile->private_disk_location, &ptr_utFile->ptr_utDirObject->public_disk_location, sizeof(ptr_utFile->ptr_utDirObject->public_disk_location)); // copy the referenced directory details
        ptr_utFile->ucFileMode = UTFAT_FILE_IS_DIR;                      // mark that the entry is not a file but a directory
        ptr_utFile->ucDrive = ptr_utFile->ptr_utDirObject->ucDrive;
    }
    return iReturn;
}

// Rename a referenced file
//
extern int utRenameFile(const CHAR *ptrFilePath, UTFILE *ptr_utFile)
{
    DIR_ENTRY_STRUCTURE_FAT32 *ptrFileEntry;
    CHAR cFileName[8 + 3 + 1];                                           // space for 8:3 format plus an NT specific byte
    const CHAR *ptrFilename = ptrFilePath;
    UTDISK *ptr_utDisk = &utDisks[ptr_utFile->ucDrive];
    if ((fnCreateNameParagraph(&ptrFilename, cFileName)) < 0) {          // fill the file name with a single paragraph from the file path
        return UTFAT_PATH_NOT_FOUND;
    }
    ptrFileEntry = (DIR_ENTRY_STRUCTURE_FAT32 *)ptr_utDisk->ptrSectorData; // the directory entry in the sector buffer
    ptrFileEntry += ptr_utFile->private_disk_location.ucDirectoryEntry;  // move to the file entry
    if (fnLoadSector(ptr_utDisk, ptr_utFile->private_disk_location.directory_location.ulSector) != UTFAT_SUCCESS) { // ensure that the directory sector is loaded
        return UTFAT_DISK_READ_ERROR;
    }

    uMemcpy(ptrFileEntry->DIR_Name, cFileName, 11);                      // add the short name
    ptrFileEntry->DIR_NTRes = cFileName[11];
    while (utCommitSector(ptr_utDisk, ptr_utFile->private_disk_location.directory_location.ulSector) == CARD_BUSY_WAIT) {} // force writeback to finalise the operation
    ptr_utDisk->ucDiskFlags &= ~WRITEBACK_BUFFER_FLAG;
    return UTFAT_SUCCESS;
}

// A close only has real significance when the file was opened as a managed file
//
extern void utCloseFile(UTFILE *ptr_utFile)
{
    if (ptr_utFile->ucFileMode & UTFAT_MANAGED_MODE) {
        utManagedFiles[ptr_utFile->iFileHandle].managed_owner = 0;       // free the managed file entry
    }
    uMemset(ptr_utFile, 0, sizeof(UTFILE));
}


// The file has changed so update all user information
//
static void fnSynchroniseManagedFile(UTFILE *ptr_utFile)
{
    int iFileHandle = 0;
    while (iFileHandle < UTMANAGED_FILE_COUNT) {
        if (utManagedFiles[iFileHandle].managed_owner != 0) {
            if (uMemcmp(&ptr_utFile->private_file_location, &utManagedFiles[iFileHandle].utManagedFile->private_file_location, sizeof(FILE_LOCATION)) == 0) {
                utManagedFiles[iFileHandle].utManagedFile->ulFileSize = ptr_utFile->ulFileSize; // update the file size
                if (utManagedFiles[iFileHandle].utManagedFile->ulFileSize < utManagedFiles[iFileHandle].utManagedFile->ulFilePosition) {
                    utManagedFiles[iFileHandle].utManagedFile->ulFilePosition = utManagedFiles[iFileHandle].utManagedFile->ulFileSize;
                }
            }
        }
        iFileHandle++;
    }
}

// Check whether the file is locked by another user
//
static int fnFileLocked(UTFILE *ptr_utFile)
{
    int iFileHandle = 0;
    while (iFileHandle < UTMANAGED_FILE_COUNT) {
        if (utManagedFiles[iFileHandle].managed_owner != 0) {
            if (uMemcmp(&ptr_utFile->private_file_location, &utManagedFiles[iFileHandle].utManagedFile->private_file_location, sizeof(FILE_LOCATION)) == 0) {
                if (utManagedFiles[iFileHandle].managed_mode & UTFAT_PROTECTED) {
                    return 1;                                            // another user is locking this file
                }
            }
        }
        iFileHandle++;
    }
    return 0;                                                            // this file is not being locked by another user
}


extern int utWriteFile(UTFILE *ptr_utFile, unsigned char *ptrBuffer, unsigned short usLength)
{
    unsigned short usWriteLength;
    unsigned short usRemainingBuffer;
    unsigned char  ucDataContent[512];
    UTDISK *ptr_utDisk = &utDisks[ptr_utFile->ucDrive];
    ptr_utFile->usLastReadWriteLength = 0;
    if (!(ptr_utFile->ucFileMode & UTFAT_OPEN_FOR_WRITE)) {
        return UTFAT_FILE_NOT_WRITEABLE;
    }
    while (usLength != 0) {
        usRemainingBuffer = (unsigned short)(ptr_utFile->ulFilePosition % 512);
        if ((usRemainingBuffer != 0) || ((usLength < 512) && (usLength < (ptr_utFile->ulFileSize - ptr_utFile->ulFilePosition)))) { // if the complete content is not being overwritten
            if (fnLoadData(ptr_utDisk, ptr_utFile->public_file_location.ulSector, ucDataContent, 0, 512) != UTFAT_SUCCESS) { // read directly to buffer
                return UTFAT_DISK_READ_ERROR;
            }
        }
        else {
            uMemset(ucDataContent, 0, sizeof(ucDataContent));
        }
        usWriteLength = usLength;
        if (usWriteLength > (512 - usRemainingBuffer)) {
            usWriteLength = (512 - usRemainingBuffer);
        }
        uMemcpy(&ucDataContent[usRemainingBuffer], ptrBuffer, usWriteLength);
        while (utCommitSectorData(ptr_utDisk, ucDataContent, ptr_utFile->public_file_location.ulSector) == CARD_BUSY_WAIT) {}
        ptrBuffer += usWriteLength;
        ptr_utFile->ulFilePosition += usWriteLength;
        if (ptr_utFile->ulFilePosition > ptr_utFile->ulFileSize) {
            ptr_utFile->ulFileSize = ptr_utFile->ulFilePosition;
        }
        ptr_utFile->usLastReadWriteLength += usWriteLength;              // the number of bytes written
        usLength -= usWriteLength;
        if ((usRemainingBuffer + usWriteLength) >= 512) {
            int iResult = fnNextSectorCreate(ptr_utDisk, &(ptr_utFile->public_file_location));
            if (iResult != UTFAT_SUCCESS) {                              // move to next sector/cluster and create additional cluster if necessary
                return iResult;
            }
        }
    }
    if (ptr_utFile->usLastReadWriteLength != 0) {                        // updata the file entry if there was a change
        DIR_ENTRY_STRUCTURE_FAT32 *ptrFileEntry = (DIR_ENTRY_STRUCTURE_FAT32 *)ptr_utDisk->ptrSectorData; // the directory entry in the sector buffer
        ptrFileEntry += ptr_utFile->private_disk_location.ucDirectoryEntry; // move to the file entry
        if (fnLoadSector(ptr_utDisk, ptr_utFile->private_disk_location.directory_location.ulSector) != UTFAT_SUCCESS) { // ensure that the directory sector is loaded
            return UTFAT_DISK_READ_ERROR;
        }
        fnSetFileInformation(ptrFileEntry, ptr_utFile->ulFileSize);
        while (utCommitSector(ptr_utDisk, ptr_utFile->private_disk_location.directory_location.ulSector) == CARD_BUSY_WAIT) {} // force writeback to finalise the operation
        ptr_utDisk->ucDiskFlags &= ~WRITEBACK_BUFFER_FLAG;
    }
    fnSynchroniseManagedFile(ptr_utFile);
    return UTFAT_SUCCESS;
}



// Change the directory location with reference to its present position
//
extern int utChangeDirectory(const CHAR *ptrDirPath, UTDIRECTORY *ptrDirObject)
{
    int iMoveUp = 0;
    unsigned short usPathTerminator = ptrDirObject->usRelativePathLocation;
    while ((*ptrDirPath == '.') && (*(ptrDirPath + 1) == '.')) {
        if (ptrDirObject->ptrDirectoryPath == 0) {                       // if no directory string path in use only referenced paths possible to always move to root
            ptrDirObject->ucDirectoryFlags |= UTDIR_TEST_FULL_PATH;
            ptrDirPath = 0;
            break;
        }

        if (usPathTerminator <= 3) {                                     // can't move up any further
            return UTFAT_PATH_NOT_FOUND;
        }
        if (iMoveUp != 0) {
            usPathTerminator--;
        }
        while ((ptrDirObject->ptrDirectoryPath[usPathTerminator] != BACK_SLASH) && (ptrDirObject->ptrDirectoryPath[usPathTerminator] != FORWARD_SLASH)) {
            usPathTerminator--;
        }
        ptrDirPath += 2;
        iMoveUp++;
        if (*ptrDirPath == 0) {
            break;
        }
        ptrDirPath++;
    }
    if (iMoveUp != 0) {
        ptrDirObject->ucDirectoryFlags |= UTDIR_TEST_FULL_PATH;          // use full path to set directory but don't move original (this flag is automatically reset)
        if (usPathTerminator > 3) {                                      // not root directory
            CHAR cTemp = ptrDirObject->ptrDirectoryPath[usPathTerminator]; // backup the location
            ptrDirObject->ptrDirectoryPath[usPathTerminator] = 0;        // temporary termination
            if (utOpenDirectory(&ptrDirObject->ptrDirectoryPath[3], ptrDirObject) != UTFAT_SUCCESS) { // try to locate the new directory
                ptrDirObject->ptrDirectoryPath[usPathTerminator] = cTemp;// correct the original path
                return UTFAT_PATH_NOT_FOUND;
            }
            ptrDirObject->ptrDirectoryPath[usPathTerminator] = cTemp;    // correct the original path
            ptrDirObject->ucDirectoryFlags |= UTDIR_TEST_FULL_PATH_TEMP; // use full path to set directory but don't move original (this flag is automatically reset). This time it is referenced to the intermediate directory
        }
        else {
            ptrDirObject->ptrDirectoryPath[3] = 0;
        }
    }
    else {
        ptrDirObject->ucDirectoryFlags |= UTDIR_TEST_REL_PATH;           // use relative path to set directory but don't move original (this flag is automatically reset).
    }
    if (utOpenDirectory(ptrDirPath, ptrDirObject) != UTFAT_SUCCESS) {    // try to locate the new directory
        return UTFAT_PATH_NOT_FOUND;
    }
    uMemcpy(&ptrDirObject->private_disk_location, &ptrDirObject->public_disk_location, sizeof(ptrDirObject->private_disk_location)); // commit the new directory location
    if (ptrDirObject->ptrDirectoryPath != 0) {
        if (*ptrDirPath != 0) {
            if (usPathTerminator > 3) {
                ptrDirObject->ptrDirectoryPath[usPathTerminator] = '\\';
            }
            if (usPathTerminator != 3) {
                usPathTerminator++;
            }
            ptrDirObject->usRelativePathLocation = (uStrcpy(&ptrDirObject->ptrDirectoryPath[usPathTerminator], ptrDirPath) - ptrDirObject->ptrDirectoryPath);
        }
        else {
            ptrDirObject->usRelativePathLocation = usPathTerminator;
            if (usPathTerminator == 2) {
                usPathTerminator = 3;
            }
            ptrDirObject->ptrDirectoryPath[usPathTerminator] = 0;
        }
    }
    return UTFAT_SUCCESS;
}

// Move through sectors and cluster chain until the end sector is reached
//
static int fnRunThroughClusters(UTDISK *ptr_utDisk, FILE_LOCATION *prtFileLocation, unsigned long ulStartPosition, unsigned long ulEndPosition)
{
    int iResult = UTFAT_SUCCESS;
    unsigned short usBytesPerSector = ptr_utDisk->utFAT.usBytesPerSector;
    ulStartPosition &= ~(usBytesPerSector - 1);                          // round down to start of present sector
    ulEndPosition &= ~(usBytesPerSector - 1);                            // round down to start of end sector
    while (ulStartPosition < ulEndPosition) {
        ulStartPosition += usBytesPerSector;
        if (fnNextSectorCreate(ptr_utDisk, prtFileLocation) != UTFAT_SUCCESS) { // if the end of the file is reached, which is at the end of a cluster, a new cluster will be created
            break;
        }
    }
    return iResult;
}

extern int utSeek(UTFILE *ptr_utFile, unsigned long ulPosition, int iSeekType)
{
    int iResult = UTFAT_SUCCESS;
    unsigned long ulNewPosition = 0;                                     // default to start of file
    UTDISK *ptr_utDisk = &utDisks[ptr_utFile->ucDrive];
    switch (iSeekType) {
    case UTFAT_SEEK_SET:                                                 // relative to the start of the file - always forwarsd
        ulNewPosition = ulPosition;
        break;
    case UTFAT_SEEK_CUR:                                                 // relative to current position
        if ((signed long)ulPosition < 0) {                               // relative backwards
            if (-ulPosition > ptr_utFile->ulFilePosition) {
                break;                                                   // move to beginning of file
            }
        }
        else {                                                           // relative forwards
            if (ulPosition > (ptr_utFile->ulFileSize - ptr_utFile->ulFilePosition)) {
                ulPosition = (ptr_utFile->ulFileSize - ptr_utFile->ulFilePosition);
            }
        }
        ulNewPosition = ptr_utFile->ulFilePosition + ulPosition;
        break;
    case UTFAT_SEEK_END:                                                 // relative to the end of the file (always backwards)
        if (ulPosition < ptr_utFile->ulFileSize) {
            ulNewPosition = (ptr_utFile->ulFileSize - ulPosition);
        }
        break;
    }
    if (ulNewPosition != ptr_utFile->ulFilePosition) {                   // update the cluster pointer
        if (ulNewPosition < ptr_utFile->ulFilePosition) {                // move back
            uMemcpy(&ptr_utFile->public_file_location, &ptr_utFile->private_file_location, sizeof(ptr_utFile->private_file_location)); // rewind to start and search from there
            ptr_utFile->ulFilePosition = 0;
        }
        iResult = fnRunThroughClusters(ptr_utDisk, &ptr_utFile->public_file_location, ptr_utFile->ulFilePosition, ulNewPosition);
        if (iResult == UTFAT_SUCCESS) {
            ptr_utFile->ulFilePosition = ulNewPosition;
        }
    }
    return iResult;
}

#ifdef UTFAT_WRITE
static int utCommitSector(UTDISK *ptr_utDisk, unsigned long ulSectorNumber)
{
    static unsigned long ulSector;
    static int iActionResult;
    switch (iMemoryOperation & _WRITING_MEMORY) {
    case _IDLE_MEMORY:
        if (!(ptr_utDisk->ucDiskFlags & HIGH_CAPACITY_SD_CARD)) {
            ulSectorNumber *= 512;                                       // convert the sector number to byte address
        }
        SET_SD_CS_LOW();
        iMemoryOperation |= _WRITING_MEMORY;
        ulSector = ulSectorNumber;
    case _WRITING_MEMORY:
        if (/*ucSectorCount == */1) {
            unsigned char ucResult;
            if ((iActionResult = fnSendSD_command(fnCreateCommand(WRITE_BLOCK_CMD24, ulSector), &ucResult, 0)) != 0) {
                return iActionResult;
            }
            if (ucResult == 0) {
                if (fnPutSector(ptr_utDisk->ptrSectorData) != UTFAT_SUCCESS) { // start writing buffer to single sector
                    fnMemoryDebugMsg("Write error\r\n");
                    return UTFAT_DISK_WRITE_ERROR;
                }
                if (!(ptr_utDisk->ucDiskFlags & HIGH_CAPACITY_SD_CARD)) {
                    ulSector /= 512;                                     // convert back to sector number from byte offset
                }
                ptr_utDisk->ulPresentSector = ulSector;                  // update the sector which is presently valid
            }
            SET_SD_CS_HIGH();
            iMemoryOperation &= ~_WRITING_MEMORY;                        // write operation has completed
        }
        else {
        }
        break;
    }
    return UTFAT_SUCCESS;
}

static int utCommitSectorData(UTDISK *ptr_utDisk, void *ptrBuffer, unsigned long ulSectorNumber)
{
    static unsigned long ulSector;
    static int iActionResult;
    switch (iMemoryOperation & _WRITING_MEMORY) {
    case _IDLE_MEMORY:
        if (!(ptr_utDisk->ucDiskFlags & HIGH_CAPACITY_SD_CARD)) {
            ulSectorNumber *= 512;                                       // convert the sector number to byte address
        }
        SET_SD_CS_LOW();
        iMemoryOperation |= _WRITING_MEMORY;
        ulSector = ulSectorNumber;
    case _WRITING_MEMORY:
        if (/*ucSectorCount == */1) {
            unsigned char ucResult;
            if ((iActionResult = fnSendSD_command(fnCreateCommand(WRITE_BLOCK_CMD24, ulSector), &ucResult, 0)) != 0) {
                return iActionResult;
            }
            if (ucResult == 0) {
                if (fnPutSector((unsigned char *)ptrBuffer) != UTFAT_SUCCESS) { // start writing buffer to single sector
                    fnMemoryDebugMsg("Write error\r\n");
                    return UTFAT_DISK_WRITE_ERROR;
                }
            }
            SET_SD_CS_HIGH();
            iMemoryOperation &= ~_WRITING_MEMORY;                        // write operation has completed
        }
        else {
        }
        break;
    }
    return UTFAT_SUCCESS;
}

static int utDeleteSector(UTDISK *ptr_utDisk, unsigned long ulSectorNumber)
{
    static unsigned long ulSector;
    static int iActionResult;
    switch (iMemoryOperation & _WRITING_MEMORY) {
    case _IDLE_MEMORY:
        if (!(ptr_utDisk->ucDiskFlags & HIGH_CAPACITY_SD_CARD)) {
            ulSectorNumber *= 512;                                       // convert the sector number to byte address
        }
        SET_SD_CS_LOW();
        iMemoryOperation |= _WRITING_MEMORY;
        ulSector = ulSectorNumber;
    case _WRITING_MEMORY:
        if (/*ucSectorCount == */1) {
            unsigned char ucResult;
            if ((iActionResult = fnSendSD_command(fnCreateCommand(WRITE_BLOCK_CMD24, ulSector), &ucResult, 0)) != 0) {
                return iActionResult;
            }
            if (ucResult == 0) {
                if (fnSetSector(0) != UTFAT_SUCCESS) {                   // start writing buffer to single sector
                    fnMemoryDebugMsg("Write error\r\n");
                    return UTFAT_DISK_WRITE_ERROR;
                }
            }
            SET_SD_CS_HIGH();
            iMemoryOperation &= ~_WRITING_MEMORY;                        // write operation has completed
        }
        else {
        }
        break;
    }
    return UTFAT_SUCCESS;
}

static int fnDeleteClusterChain(unsigned long ulClusterStart, unsigned char ucDrive, int iDestroyClusters)
{
    UTDISK *ptr_utDisk = &utDisks[ucDrive];
    unsigned long ulCluster = (((ulClusterStart - ptr_utDisk->ulLogicalBaseAddress)/ptr_utDisk->utFAT.ucSectorsPerCluster) + ptr_utDisk->ulDirectoryBase);
    unsigned long ulClusterSector = (ptr_utDisk->utFAT.ulFAT_start + (ulCluster >> 7)); // section where the FAT responsible for this cluster resides
    unsigned long ulNextCluster, ulNextClusterSector;
    unsigned long ulRemovedClusters = 0;
    unsigned char ucClusterEntry = (unsigned char)(ulCluster & 0x7f);
    unsigned long ulSectorContent[512/sizeof(signed long)];
    if ((utReadDiskData(ptr_utDisk, ulClusterSector, ulSectorContent, 0, 512)) != UTFAT_SUCCESS) { // read a FAT sector containing the cluster information
        return UTFAT_DISK_READ_ERROR;
    }
    while (1) {
        ulNextCluster = ulSectorContent[ucClusterEntry];                 // read the next entry
        if (((ulNextCluster & CLUSTER_MASK) == CLUSTER_MASK) || (ulNextCluster <= ptr_utDisk->ulDirectoryBase)) { // check whether the end of the cluster chain has been reached (if the next cluster is 0, 1 or 2 it is ignored since it must be corrupt - 0, 1 are not used and 2 is always the start of the root directory!)
            int iResult = UTFAT_SUCCESS;
            unsigned char ucFatCopies = 0;
            if (iDestroyClusters == 0) {                                 // if the file's cluster is to be reused and only occupied one cluster, don't delete it
                break;
            }
            else {
                ulSectorContent[ucClusterEntry] = 0;                     // delete the previous entry
                if (ptr_utDisk->ucDiskFlags & FSINFO_VALID) {
                    ptr_utDisk->utFileInfo.ulFreeClusterCount += ++ulRemovedClusters;
                    ptr_utDisk->ucDiskFlags |= WRITEBACK_INFO_FLAG;      // mark that the info block information has changed
                }
            }
            while (ucFatCopies < ptr_utDisk->utFAT.ucNumberOfFATs) {
                while ((iResult = utCommitSectorData(ptr_utDisk, ulSectorContent, (ulClusterSector + (ucFatCopies * ptr_utDisk->utFAT.ulFatSize)))) == CARD_BUSY_WAIT) {}
                ucFatCopies++;
            }
            return iResult;
        }
        if (iDestroyClusters == 0) {                                     // if the file's cluster is to be reused
            ulSectorContent[ucClusterEntry] = CLUSTER_MASK;              // mark the end of the cluster chain
            iDestroyClusters = 1;                                        // from here on, destroy rest of chain
        }
        else {
            ulSectorContent[ucClusterEntry] = 0;                         // delete the previous entry
            ulRemovedClusters++;
        }
        ulNextClusterSector = (ptr_utDisk->utFAT.ulFAT_start + (ulNextCluster >> 7));
        ucClusterEntry = (unsigned char)(ulNextCluster & 0x7f);
        if (ulNextClusterSector != ulClusterSector) {                    // moving to a new sector so update the FAT
            unsigned char ucFatCopies = 0;
            while (ucFatCopies < ptr_utDisk->utFAT.ucNumberOfFATs) {
                while (utCommitSectorData(ptr_utDisk, ulSectorContent, (ulClusterSector + (ucFatCopies * ptr_utDisk->utFAT.ulFatSize))) == CARD_BUSY_WAIT) {}
                ucFatCopies++;
            }
            while (utCommitSectorData(ptr_utDisk, ulSectorContent, ulClusterSector) == CARD_BUSY_WAIT) {}
            ulClusterSector = ulNextClusterSector;
            if ((utReadDiskData(ptr_utDisk, ulClusterSector, ulSectorContent, 0, 512)) != UTFAT_SUCCESS) { // read a FAT sector containing the cluster information
                return UTFAT_DISK_READ_ERROR;
            }
        }
    }
    return UTFAT_SUCCESS;
}

static int fnDeleteFileContent(UTFILE *ptr_utFile, UTDISK *ptr_utDisk, int iDestroyClusters)
{
    int iResult;
    ptr_utDisk->ucDiskFlags |= WRITEBACK_BUFFER_FLAG;                    // mark that the content has been changed and must be written back when the buffer is next used
#ifdef UTFAT_LFN_READ                                                    // if long file names are in use, delete also possible long file name
  //while (1) {
  //    fnPreviousDirectoryEntry(&utDisks[ptrDirObject->ucDrive], &(utFile.private_disk_location));
  //}
#endif    
    iResult = fnDeleteClusterChain(ptr_utFile->private_file_location.ulSector, ptr_utDisk->ucDriveNumber, iDestroyClusters);
    if (ptr_utDisk->ucDiskFlags & WRITEBACK_BUFFER_FLAG) {
        if (iDestroyClusters == 0) {                                     // the file is to be re-used so set length to zero
            fnSetFileInformation(ptr_utFile->ptr_utDirObject->ptrEntryStructure, 0);// set the file size to zero
        }
        while (utCommitSector(ptr_utDisk, ptr_utDisk->ulPresentSector) == CARD_BUSY_WAIT) {} // force writeback to finalise the operation
        ptr_utDisk->ucDiskFlags &= ~WRITEBACK_BUFFER_FLAG;
    }
    return fnCommitInfoChanges(ptr_utDisk);
}

extern int utDeleteFile(const CHAR *ptrFilePath, UTDIRECTORY *ptrDirObject)
{
    int iResult;
    UTFILE utFile;
    UTDISK *ptr_utDisk = &utDisks[ptrDirObject->ucDrive];
    utFile.ptr_utDirObject = ptrDirObject;
    ptrDirObject->ucDirectoryFlags |= (UTDIR_TEST_REL_PATH);
    iResult = utOpenFile(ptrFilePath, &utFile, UTFAT_OPEN_FOR_DELETE);
    if (iResult == UTFAT_SUCCESS) {                                      // directory has been located
        DISK_LOCATION *ptrDirContent = &utFile.private_disk_location;
        DIR_ENTRY_STRUCTURE_FAT32 *ptrDirEntry;
        uMemcpy(&utFile.private_file_location, &utFile.private_disk_location.directory_location, sizeof(utFile.private_disk_location.directory_location));
        do {
            if (fnLoadSector(ptr_utDisk, utFile.private_disk_location.directory_location.ulSector) != UTFAT_SUCCESS) {
                return UTFAT_DISK_READ_ERROR;
            }
            ptrDirEntry = (DIR_ENTRY_STRUCTURE_FAT32 *)ptr_utDisk->ptrSectorData; // the directory entry in the sector buffer
            ptrDirEntry += ptrDirContent->ucDirectoryEntry;              // move to the present entry
            if (ptrDirEntry->DIR_Name[0] == 0) {
                break;                                                   // directory is empty so allow delete
            }
            else if (ptrDirEntry->DIR_Name[0] != DIR_NAME_FREE) {
                return UTFAT_DIR_NOT_EMPTY;
            }
            if (fnNextDirectoryEntry(ptr_utDisk, ptrDirContent) == UTFAT_DIRECTORY_AREA_EXHAUSTED) { // move to the next entry
                return UTFAT_DIRECTORY_AREA_EXHAUSTED;
            }
        } while (1);
        ptrDirObject->ucDirectoryFlags |= (UTDIR_TEST_REL_PATH | UTDIR_DIR_AS_FILE);
        iResult = utOpenFile(ptrFilePath, &utFile, UTFAT_OPEN_FOR_DELETE);
        if (iResult == UTFAT_SUCCESS) {
            utFile.ptr_utDirObject->ptrEntryStructure->DIR_Name[0] = DIR_NAME_FREE; // delete the (short) file name
            return (fnDeleteFileContent(&utFile, &utDisks[ptrDirObject->ucDrive], 1));
        }
    }
    else if (iResult == UTFAT_PATH_IS_FILE) {                            // if file has been located
        utFile.ptr_utDirObject->ptrEntryStructure->DIR_Name[0] = DIR_NAME_FREE; // delete the (short) file name
        return (fnDeleteFileContent(&utFile, &utDisks[ptrDirObject->ucDrive], 1));
    }
    return iResult;
}

// Make a new directory - don't allow if a directory of the same name already exists
//
extern int utMakeDirectory(const CHAR *ptrDirPath, UTDIRECTORY *ptrDirObject)
{
    int iReturn;
    DISK_LOCATION *ptrDiskLocation = &(ptrDirObject->public_disk_location);
    DISK_LOCATION found_location;
    UTDISK *ptr_utDisk = &utDisks[ptrDirObject->ucDrive];
    DIR_ENTRY_STRUCTURE_FAT32 *ptrFoundEntry;
    const CHAR *ptrDirname = ptrDirPath;
    CHAR cFileName[8 + 3 + 1];                                           // space for 8:3 format plus an NT specific byte

    ptrDirObject->ucDirectoryFlags |= (UTDIR_SET_START);                 // {6} set to start of lowest directory when file not found
    if ((iReturn = utOpenDirectory(ptrDirPath, ptrDirObject)) != UTFAT_FILE_NOT_FOUND) {
        return UTFAT_PATH_NOT_FOUND;
    }
    do {
        iReturn = fnCreateNameParagraph(&ptrDirname, cFileName);         // run down the path until the directory name is loaded
    } while ((iReturn != FULLY_QUALIFIED_SHORT_NAME) && (iReturn != FULLY_QUALIFIED_LONG_NAME));

    found_location.directory_location.ulCluster = 0;

    while (1) {
        if (fnLoadSector(ptr_utDisk, ptrDiskLocation->directory_location.ulSector) != UTFAT_SUCCESS) {
            return UTFAT_DISK_READ_ERROR;
        }
        ptrFoundEntry = (DIR_ENTRY_STRUCTURE_FAT32 *)ptr_utDisk->ptrSectorData; // the directory entry in the sector buffer
        ptrFoundEntry += ptrDiskLocation->ucDirectoryEntry;              // move to the present entry
        if (ptrFoundEntry->DIR_Name[0] == 0) {                           // end of root entries found
            unsigned long ulFreeCluster;
            FILE_LOCATION *ptr_newDir;
            if (found_location.directory_location.ulCluster != 0) {      // fill a deleted location which was found during the check of existing directory
                ptr_newDir = &found_location.directory_location;
                if (fnLoadSector(ptr_utDisk, ptr_newDir->ulSector) != UTFAT_SUCCESS) {
                    return UTFAT_DISK_READ_ERROR;
                }
                ptrFoundEntry = (DIR_ENTRY_STRUCTURE_FAT32 *)ptr_utDisk->ptrSectorData; // the directory entry in the sector buffer
                ptrFoundEntry += found_location.ucDirectoryEntry;        // move to the present entry
            }
            else {
                ptr_newDir = &ptrDiskLocation->directory_location;
            }
            ulFreeCluster = fnAllocateCluster(ptr_utDisk, ptr_newDir->ulCluster, INITIALISE_DIR_CLUSTER);
            uMemcpy(ptrFoundEntry->DIR_Name, cFileName, 11);             // add the short name
            ptrFoundEntry->DIR_NTRes = cFileName[11];
            fnAddEntry(ptrFoundEntry, ulFreeCluster, DIR_ATTR_DIRECTORY);
            while (utCommitSector(ptr_utDisk, ptr_newDir->ulSector) == CARD_BUSY_WAIT) {} // force writeback to finalise the operation
            ptr_utDisk->ucDiskFlags &= ~WRITEBACK_BUFFER_FLAG;
            return (fnCommitInfoChanges(ptr_utDisk));
        }
        else if (ptrFoundEntry->DIR_Name[0] == DIR_NAME_FREE) {          // a free deleted entry found
            if (found_location.directory_location.ulCluster == 0) {
                uMemcpy(&found_location, ptrDiskLocation, sizeof(found_location)); // remember the entry location
            }
        }
        else {                                                           // a valid entry found
            if (ptrFoundEntry->DIR_Attr == DIR_ATTR_DIRECTORY) {         // if it is a directory
                if (!(uMemcmp(ptrFoundEntry->DIR_Name, cFileName, sizeof(ptrFoundEntry->DIR_Name)))) { // check that the directory doesn't already exist
                    return UTFAT_DIRECTORY_EXISTS_ALREADY;               // return an error since it is not allowed to have two directories with the same name
                }
            }
        }
        if ((fnNextDirectoryEntry(ptr_utDisk, ptrDiskLocation) != UTFAT_SUCCESS) && (found_location.directory_location.ulCluster == 0)) { // move to next directory entry
            int iResult;
            iResult = fnDirectorySectorCreate(ptr_utDisk, &ptrDiskLocation->directory_location); // move to next sector/cluster and create additional cluster if necessary
            if (iResult != UTFAT_SUCCESS) {
                return iResult;
            }
        }
    }
    return UTFAT_SUCCESS;
}

extern int utReFormat(const unsigned char ucDrive, const CHAR *cVolumeLabel)
{
    iMemoryState = SD_STATE_FORMATTING_DISK_1;
    iMemoryOperation = 0;
    uMemcpy(utDisks[ucDrive].cVolumeLabel, cVolumeLabel, 11);
    uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);                      // schedule the task to start formatting
    return UTFAT_SUCCESS;
}

extern int utFormat(const unsigned char ucDrive, const CHAR *cVolumeLabel)
{
    if (iMemoryState != SD_NOT_FORMATTED) {
        return UTFAT_DISK_NOT_READY;
    }
    return utReFormat(ucDrive, cVolumeLabel);
}
#endif


// This routine is called to count the free clusters and update the disk information
//
extern int utFreeClusters(unsigned char ucDisk, UTASK_TASK owner_task)
{
    if (cluster_task != 0) {                                             // cluster count already in progress
        return -1;
    }
    ulActiveFreeClusterCount = 0;
    ulClusterSectorCheck = (utDisks[ucDisk].utFAT.ulFAT_start + (utDisks[ucDisk].utFAT.ulFatSize - 1));
    iMemoryOperation |= _COUNTING_CLUSTERS;
    cluster_task = owner_task;
    uTaskerStateChange(OWN_TASK, UTASKER_ACTIVATE);                      // start cluster counting operation
    return 0;
}


// Debug routines
//
extern int fnReadSector(unsigned char ucDisk, unsigned char *ptrBuffer, unsigned long ulSectorNumber)
{
    return (fnLoadData(&utDisks[ucDisk], ulSectorNumber, ptrBuffer, 0, 512));
}

extern UTDISK *fnGetDiskInfo(unsigned char ucDisk)
{
    return (&utDisks[ucDisk]);
}



#else
extern void fnMassStorage(TTASKTABLE *ptrTaskTable)
{
}
#endif
